

class GutenbergBook:

    def __init__(self, author_name: str = "", author_birth: int = 0, author_death: int = 0, title: str = "",
                 lang: str = "", genre: str = "", subject: str = "", num_chars: int = 0, num_words: int = 0,
                 num_sentences: int = 0, num_difficult_words: int = 0, url: str = "", downloads: int = 0):
            self._author_name = author_name
            self._author_birth = author_birth
            self.author_death = author_death
            self._title = title
            self._lang = lang
            self._genre = genre
            self._subject = subject
            self._num_chars = num_chars
            self._num_words = num_words
            self._num_sentences = num_sentences
            self._num_difficult_words = num_difficult_words
            self._url = url
            self._downloads = downloads

    @property
    def name(self):
        return self._author_name

    @name.setter
    def name(self, n):
        self._author_name = n

    @property
    def birth(self):
        return self._author_birth

    @birth.setter
    def birth(self, b):
        self._author_birth = b

    @property
    def death(self):
        return self._author_death

    @death.setter
    def death(self, d):
        self._author_death = d

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, t):
        self._title = t

    @property
    def lang(self):
        return self._lang

    @lang.setter
    def lang(self, l):
        self._lang = l

    @property
    def genre(self):
        return self._genre

    @genre.setter
    def genre(self, g):
        self._genre = g

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, s):
        self._subject = s

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, u):
        self._url = u

    @property
    def num_chars(self):
        return self._num_chars

    @num_chars.setter
    def num_chars(self, n):
        self._num_chars = n

    @property
    def num_words(self):
        return self._num_words

    @num_words.setter
    def num_words(self, w):
        self._num_words = w

    @property
    def num_sentences(self):
        return self._num_sentences

    @num_sentences.setter
    def num_sentences(self, n):
        self._num_sentences = n

    @property
    def num_difficult_words(self):
        return self._num_difficult_words

    @num_difficult_words.setter
    def num_difficult_words(self, w):
        self._num_difficult_words = w

    @property
    def downloads(self):
        return self._downloads

    @downloads.setter
    def downloads(self, d):
        self._downloads = d




from bridges.bst_element import *
##
 #  @brief The This class can be used to create K-d Tree  elements, derived
 #  from BSTElement. K-D trees can be thought of as the spatial equivalent
 #  binary search trees and operate on multiple dimensions (2D and 3D are
 #  most common). These trees serve as a representation of the underlying
 #  geometrically defined spaces. Specialized versions of these trees
 #  include quadtrees and octrees, which subdivide into equal sized quadrants or
 #  octants at each level, respectively.
 #  This class extends the BSTElement class by adding a dimension property.
 #  It also includes a thickness property for displaying the partitioning
 #  lines generated by the convex decomposition.
 #
 # @author Kalpathi Subramanian, Matthew McQuaigue
 # @date 12/26/18, 7/23/19
 #
 # \sa Kd Tree tutorial, http://bridgesuncc.github.io/tutorials/KdTree.html
 #

class KDTreeElement(BSTElement):

    def __init__(self, **kwargs):
        """
        Constructor for a kdTreeElement with the provided value, label, key, left and right kdTree elements.
        The defaults will be used if not provided
        Args:
            (object) key: The key for the ordering
            val: the data this element will hold
            (str) label: the label to show during visualization
            left: the left kdTree element
            right:  the right kdTree element
            (float) th: thickness of KDTree partition lines
            (int) dim: partition dimension
        Returns:
            None
        """
        if 'val' in kwargs:
            if 'label' in kwargs:
                if 'left' and 'right' in kwargs:
                    super(KDTreeElement, self).__init__(e = kwargs['val'], label = kwargs['label'],
                                                        left = kwargs['left'], right = kwargs['right'])
                else:
                    super(KDTreeElement, self).__init__(e=kwargs['val'], label=kwargs['label'])
            else:
                if 'left' and 'right' in kwargs:
                    super(KDTreeElement, self).__init__(e=kwargs['val'], left=kwargs['left'], right=kwargs['right'])
                else:
                    super(KDTreeElement, self).__init__(e=kwargs['val'])
        else:
            super(KDTreeElement, self).__init__()
        if 'key' in kwargs:
            super(KDTreeElement, self.__class__).key.fset(self, kwargs['key'])
        if 'th' in kwargs:
            self._thickness = kwargs['th']
        else:
            self._thickness = 0.0
        if 'dim' in kwargs:
            self._dimension = kwargs['dim']
        else:
            self._dimension = 0

    def get_data_structure_type(self) -> str:
        """
        Getter for the data structure type
        Returns:
            str: of the the data structure type
        """
        return "KdTree"

    @property
    def dimension(self) -> int:
        """
        Getter for the dimensions of the partition lines
        Returns:
            int: dimension (0, 1, 2, etc)
        """
        return self._dimension

    @dimension.setter
    def dimension(self, dim: int) -> None:
        """
        Setter for the dimension of th partitioning at this tree node
        Args:
            (int) dim: dimension value to set
        Returns:
            None
        """
        self._dimension = dim

    @property
    def thickness(self) -> float:
        """
        Getter for the thickness to the KDTree partitions
        Returns:
            float: the thickness
        """
        return self._thickness

    @thickness.setter
    def thickness(self, th) -> None:
        """
        Setter for the thickness of the KDTree partitions - can be used in the visualization
        Args:
            (float) th: thickness of partitions
        Returns:
            None
        """
        self._thickness = th

    @property
    def left(self):
        """
        Getter for the left child of the tree element
        Returns:
            KdTreeElement: the left child
        """
        return super(KDTreeElement, self).left

    @left.setter
    def left(self, l):
        self.set_child(0, l)

    @property
    def right(self):
        """
        Getter for the right child of the tree element
        Returns:
            KdTreeElement: the right child
        """
        return super(KDTreeElement, self).right

    @right.setter
    def right(self, r):
        self.set_child(1, r)

    def get_element_representation(self):
        orig_json_str = super(KDTreeElement, self).get_element_representation()
        kdt_dict = {
            "dimension": str(self.dimension),
            "thickness": str(self.thickness)
        }
        orig_json_str.update(kdt_dict)
        return orig_json_str




import datetime
import time


class EarthquakeUSGS:
    """
	Class that hold earthquake data, for use with USGIS retrieved quake data.
    Class that holds earthquake USGIS data. BRIDGES uses scripts to continually monitor
	USGIS site (tweets) and retrieve the latest quake data for use in student projects.
	Kalpathi Subramanian, 2/18/18
    """

    def __set_time_from_unix_timestamp(self, tm):
        epoch_time = int(tm)
        eq_time = epoch_time / 1000
        eqt = time.gmtime(eq_time)
        self._time = time.strftime("%Y-%m-%d %H:%M:%S", eqt)

    def __init__(self, magnitude=None, longit=None, latit=None, location=None, title=None, url=None, time=None):
        self._time = int()
        if magnitude is not None:
            self._magnitude = magnitude
        else:
            self._magnitude = 0.0
        if longit is not None:
            self._longit = longit
        else:
            self._longit = 0.0
        if latit is not None:
            self._latit = latit
        else:
            self._latit = 0.0
        if location is not None:
            self._location = location
        else:
            self._location = ""
        if title is not None:
            self._title = title
        else:
            self._title = ""
        if url is not None:
            self._url = url
        else:
            self._url = ""
        if time is not None:
            self.time = time

    @property
    def time(self):
        """
        Getter for the time of the earthquakle
        Returns:
            time of earthquake
        """
        return self._time

    @time.setter
    def time(self, tm) -> None:
        """
        Set epoch time
        Args:
            tm: time as epoch
        """
        self.__set_time_from_unix_timestamp(tm)

    @property
    def latit(self) -> float:
        """
        Get latitude of quake
        Returns:
            latitude of the quake
        """
        return self._latit

    @latit.setter
    def latit(self, latit: float) -> None:
        """
        Setter for the latitude
        Args:
            latit: latitiude to set
        """
        self._latit = latit

    @property
    def longit(self) -> float:
        """
        Getter for the longitude of quake
        Returns:
            logitude
        """
        return self._longit

    @longit.setter
    def longit(self, longit: float) -> None:
        """
        Setter for the longitutde of the quake
        Args:
            longit: logitude to set
        """
        self._longit = longit

    @property
    def location(self) -> str:
        """
        Getter for the location of quake (typically a city or something of the sort)
        Returns:
            location
        """
        return self._location

    @location.setter
    def location(self, location: str):
        """
        Setter for the location of the quake
        Args:
            location: location to set
        """
        self._location = location

    @property
    def title(self) -> str:
        """
        Getter for the title of quake (typically a one line description)
        Returns:
            title
        """
        return self._title

    @title.setter
    def title(self, title: str):
        """
        Setter for the title of the quake
        Args:
            title: title to set
        """
        self._title = title

    @property
    def url(self) -> str:
        """
        Getter for the url of quake
        Returns:
            url
        """
        return self._url

    @url.setter
    def url(self, url: str):
        """
        Setter for the url of the quake
        Args:
            url: url to set
        """
        self._url = url

    @property
    def magnitude(self) -> float:
        """
        Getter for the magnitude of quake on the scale on Richter scale
        Returns:
            magnitude
        """
        return self._magnitude

    @magnitude.setter
    def magnitude(self, magnitude: float):
        """
        Setter for the magnitude of the quake
        Args:
            magnitude: magnitude to set
        """
        self._magnitude = magnitude




class Game:

    def __init__(self, title: str = "", platform: str = "", rating: float = 0.0, genre: str = ""):
            self._title = title
            self._platform = platform
            self._rating = rating
            self._genre = genre

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, t):
        self._title = t

    @property
    def platform(self):
        return self._platform

    @platform.setter
    def platform(self, p):
        self._platform = p

    @property
    def rating(self):
        return self._rating

    @rating.setter
    def rating(self, r):
        self._rating = r

    @property
    def genre(self):
        return self._genre

    @genre.setter
    def genre(self, g):
        self._genre = g



from bridges.polyline import *


class Polygon(Polyline):

    def __init__(self, pts = None):
        if pts is not None:
            super(Polygon, self).__init__(pts = pts)
        else:
            super(Polygon, self).__init__()
        self.shape_type = "polygon"

    def get_name(self):
        return "polygon"



#!/usr/bin/env python
from decimal import Decimal
from bridges.element import *
from bridges.color import *
from decimal import Decimal
##
# This class is used to store the visualization elements on the for the bridges
# Visualization, including the color, shape, opacity, and size of the node.<p>
#
# <b>Supported named colors: CSS colors. 
# Check https://drafts.csswg.org/css-color-3/#svg-color</b>: <p>
#
# <b> Color by RGBA Specification :</b>  Range: 0-255 for each component <p>
#
# <b> Thickness: </b> Range : 0.0-50.0
#
# <b> Opacity: </b> Range (0.0-1.0) </p>
#
# <b> Supported Shapes: </b> "circle", "square", "diamond", "cross", "triangle", "star", "wye". 
#
# Objects of this class are stored as properties of all Element subclasses.
# Generally, you will manipulating the ElementVisualizer returned from the
# Element getVisualizer() method, and then call the setVisualizer() method on
# the Element after changes have been made.
#
# @author Matthew Mcquaigue
#
# @date 2018, 6/24/19
#
class ElementVisualizer():

    def __init__(self, color="green", shape ="circle", size=10.0, opacity=1.0):
        """
        ElementVisualizer constuctor
        kwargs:
            (str) color: color for this element
            (str) shape: the shape of the element
            (float) size: the element size
            (float) opacity: the element opacity
        Returns:
            None
        """
        self.prop = dict()
        self._color = Color(70, 130, 180, 1.0)
        if color is not "green":
            self.color = color
        self._shape = shape
        self._size = size
        self._opacity = opacity
        self._locationX = Decimal("Infinity")
        self._locationY = Decimal("Infinity")
        self.prop['color'] = ["70", "130", "180", "1.0"]
        self.prop["opacity"] = "1.0"
        self.prop["size"] = "10.0"
        self.prop["shape"] = "circle"
        self.prop["key"] = ""
        self.prop["locationX"] = Decimal("Infinity")
        self.prop["locationY"] = Decimal("Infinity")


    @property
    def size(self) -> float:
        """
        Getter for the elements size
        Returns:
            float: the size
        """
        return self._size

    @size.setter
    def size(self, size) -> None:
        """
        Setter for the size of the element
        Args:
            (float) size: the elements desired size
        Returns:
            None
        """
        self.prop["size"] = str(size)
        self._size = size

    @property
    def color(self) -> Color:
        """
        Getter for the color of the element in the bridges visualization
        Returns:
            Color: Color object representing the color of the element
        """
        return self._color

    @color.setter
    def color(self, *args, **kwargs) -> None:
        """
        Setter for the color of the element in the bridges visualization
        Args:
            (optional) list: requires either 3 ints 0-255 for RGB and an optional float 0.0-1.0 for alpha EX: color = [0, 255, 0, 1.0]
            (optional) str: string representing the element color. from web colors: https://developer.mozilla.org/en-US/docs/Web/CSS/color_value
        Returns:
            None
        Raises:
            ValueError: if the color name provided is not available
        """
        self._color = Color(*args, **kwargs)

    @property
    def shape(self) -> str:
        """
        Getter for the shape of the element
        Returns:
            str: reperesenting the type of shape
        """
        return self._shape

    @shape.setter
    def shape(self, a_shape):
        """
        Setter for the shape of the element
        Args:
            (str) a_shape: the name of shape for element
        """
        a_shape = a_shape.lower()
        self.prop["shape"] = a_shape
        self._shape = a_shape

    @property
    def opacity(self) -> float:
        """
        Getter for the opacity of the element
        Returns:
            float: representing the opacity
        """
        return self.color.alpha

    @opacity.setter
    def opacity(self, opacity) -> None:
        """
        Setter for the opacity of the element
        Args:
            (float) opacity: the opacity to be applied
        Returns:
            None
        """
        self.prop["opacity"] = Decimal(opacity)
        self.color.alpha = opacity
        self._opacity = opacity

    def set_location(self, x, y):
        """
        Setter for the location of the element
        Args:
            (int) x: x location
            (int) y: y location
        Returns:
            None
        """
        self.prop['locationX'] = x
        self.prop['locationY'] = y
        self._locationX = x
        self._locationY = y

    @property
    def location_x(self):
        """
        Getter for the X location of element
        Returns:
            int: as the x location
        """
        return self._locationX

    @property
    def location_y(self):
        """
        Getter for the y location of the element
        Returns:
            int: as the y position
        """
        return self._locationY



#!/usr/bin/env python
from typing import Any, Union

from bridges.sl_element import *
from bridges.edge import *
import traceback


##
#
#    @brief The GraphAdjList class can be used to represent adjacency list based
#        graphs in BRIDGES
#
#    The GraphAdjList class can be used to represent adjacency list based  graphs
#    in BRIDGES; it takes 2 generic parameters: (1) K, which is an orderable
#    key value used in accessing vertices (in constant time) using a hashmap. This
#    permits data sets that need to be accessed by keys that are strings, and
#    (2) E, an application defined type, and used in the Edge representation.
#    The class is simply a wrapper  around the Java Hashmap class
#    and, thus, derives all its operations from it.
#    BRIDGES provides methods to visualize the graph  and its contents.
#
#    The vertices of the graph are held in a Java hashmap, for near constant time access;
#    this lets us use strings or integral ids for vertices. The adjacency lists,
#    also a Java hashmap  are built for each vertex and contain the edge (terminating
#    vertex id, weight) in the Edge structure, defined separately. Adjacency lists
#    are singly linked lists using the BRIDGES SLelement.
#
#    Convenience methods are provided to add vertices and edges to the graph as well as
#    retrieve the adjacency list of a vertex, given its id.
#
#   @author Matthew Mcquaigue
#    @date 2018,  7/23/19
#
#    \sa graph adjacency list tutorial, http://bridgesuncc.github.io/tutorials/Graph_AL.html
#
class GraphAdjList:
    LargeGraphVertSize = 1000
    force_large_viz = False
    force_small_viz = False

    def __init__(self) -> None:
        """
        Constructor for a graph adj list
        Returns:
            None
        """
        self._vertices = dict()
        self.adj_list = dict()

    def get_data_structure_type(self) -> str:
        """
        Getter for the data structure type
        Returns:
            str: representing the type
        """
        if (GraphAdjList.force_large_viz == True or (GraphAdjList.force_small_viz == False and
                                                     self.LargeGraphVertSize < len(self.vertices) and
                                                     self.are_all_vertices_located())):
            return "largegraph"
        return "GraphAdjacencyList"

    def add_vertex(self, id: str,
                   label: str = None,
                   data: Any = None,
                   color: Union[str, Color] = "blue",
                   opacity: float = 1.0,
                   original: Element = None
                   ) -> Element:
        """
        Adds a new vertex to the graph, initializes the adjacency
        list; user is responsible for checking if the vertex already exists.
        This method will replace the value for this key
        Args:
            id: the vertex id
            label: vertex label
            data: the vertex data
            color: color of vertex
            opacity: opacity of vertex
            original: an existing Element instance to copy, this will nullify other arguments
        Returns:
            newly created Vertex
        """
        #  note: it is the user's responsibility to  check
        #  for duplicate vertices
        self.vertices[id] = Element(val=data, color=color, opacity=opacity, original=original)
        self.vertices.get(id).label = str(label) if label is not None else str(id)
        self.adj_list[id] = None

        return self.vertices[id]

    def add_edge(self,
                 src: Union[str, Element],
                 dest: Union[str, Element],
                 label: str = None,
                 data: any = None,
                 color: Union[str, Color] = "blue",
                 thickness: float = 1.0,
                 opacity: float = 1.0
                 ) -> Edge:
        """
        Adds a new edge to the graph, adds it to that vertex's
        adjacency list; user is responsible for checking if the
        vertex already exists. This version assumes a default edge
        weight of 1.
        Args:
            src: source vertex of edge
            dest: destination  vertex of edge
            label: label of edge,
            color: color of edge,
            thickness: thickness of edge,
            opacity: opacity of edge,
            data: data the edge will hold
        Returns:
            newly created Edge
        Rasies:
            ValueError: if the src and dest vertices do not exist
        """
        source_id = self._resolve_id(src)
        dest_id = self._resolve_id(dest)
        if source_id is None or dest_id is None:
            raise ValueError("Vertex " + src + " or " + dest +
                             " does not exist! Add the vertex before creating the edge.")
        try:
            if self.vertices.get(source_id) is None or self.vertices.get(dest_id) is None:
                raise ValueError("Vertex " + src + " or " + dest +
                                 " does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self.adj_list[source_id] = SLelement(e=Edge(source_id, dest_id, data), next=self.adj_list.get(src), label=label,
                                       color=color, thickness=thickness, opacity=opacity)

        return self.adj_list[source_id].value

    def set_vertex_data(self, src, vertex_data) -> None:
        """
        Set for the data at a given vertex
        Args:
            src: the source vertex
            vertex_data: The data for the vertex
        Returns:
            None
        Raises:
             ValueError: if the source vertex doesnt exist
        """
        try:
            if self.vertices[src] is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self.vertices[src].value = vertex_data

    def get_vertex_data(self, src):
        try:
            if self.vertices[src] is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return self.vertices[src].value

    def set_edge_data(self, src, dest, edge_data):
        try:
            if self.vertices[src] is None or self.vertices[dest] is None:
                raise ValueError("Vertex " + src + " or " + dest + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        sle = self.adj_list[src]
        while sle is not None:
            edge_dest = sle.value.tov
            if edge_dest == dest:
                if edge_data is not None:
                    sle.value.set_edge_data(edge_data)
                    return
                else:
                    return sle.value.edge_data
            sle = sle.next
        if sle is None:
            raise ValueError("VEdge from " + src + " to " + dest + "does not exist!")

    @property
    def vertices(self) -> dict:
        """
        Getter for the graph nodes
        Returns:
            dict
        """
        return self._vertices

    def get_vertex(self, key):
        """
        Getter for a specific vertex in the dictionary of vertices
        Args;
            key: The associated key for the vertex
        Returns:
            vertex
        """
        return self.vertices.get(key)

    def get_adjacency_list(self, vertex=None):
        """
        Gets the adjacency list
        Args:
            vertex: input  vertex 
        Returns:
            list : adjacency list of this vertex
        """
        if vertex is not None:
            return self.adj_list.get(vertex)
        else:
            return self.adj_list

    def key_set(self):
        return self.vertices.keys()

    def value_set(self):
        return self.vertices.values()

    def out_going_edge_set_of(self, k):
        return SLelement.list_helper(start=self.get_adjacency_list(k))

    def get_edge_data(self, src, dest):
        sle = self.adj_list[src]
        while sle is not None:
            ed = sle.value
            if ed.destination == dest:
                return ed.get_edge_data()
            sle = sle.next

    def get_edge(self, src: Union[str, Element], dest: Union[str, Element]):
        source_id = self._resolve_id(src)
        dest_id = self._resolve_id(dest)
        if source_id is None or dest_id is None:
            raise ValueError("Vertex " + src + " or " + dest +
                             " does not exist! Add the vertex before creating the edge.")

        sle = self.adj_list[source_id]
        while sle is not None:
            ed = sle.value
            if ed.destination == dest_id:
                return ed
            sle = sle.next

    def _resolve_id(self, key_or_element: Union[str, Element]) -> Union[str, None]:
        if type(key_or_element) is str:
            return key_or_element
        else:
            return [key for key, value in self.vertices.items() if value == key_or_element].pop()

    def are_all_vertices_located(self):
        for element in self.vertices.items():
            el = element[1]
            elvis = el.visualizer
            if(elvis.location_x == float('inf') or elvis.location_y == float('inf')):
                return False
        return True

    def force_large_visualization(self, f):
        if f:
            GraphAdjList.force_large_viz = True
            GraphAdjList.force_small_viz = False
        else:
            GraphAdjList.force_large_viz = False

    def force_small_visualization(self, f):
        if f:
            GraphAdjList.force_small_viz = True
            GraphAdjList.force_large_viz = False
        else:
            GraphAdjList.force_small_viz = False


    ##
    #
    #     This is a convenience method to simplify access to the link visualizer;
    #     the method assumes the vertex names point to existing vertices, else an exception
    #     is thrown
    #
    #   @param src - source vertex of edge
    #   @param dest - destination vertex of edge
    #
    def get_link_visualizer(self, src, dest):
        #  get the source and destination vertex elements
        #  and check to see if they exist
        v1 = self.vertices.get(src)
        v2 = self.vertices.get(dest)
        try:
            if v1 is None or v2 is None:
                raise ValueError(
                    "Vertex " + src + " or " + dest + " does not exist! First add the vertices to the graph.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return v1.get_link_visualizer(v2)

    ##
    #
    #    This is a convenience method to simplify access to the element visualizer;
    #    the method assumes the vertex name points to an existing vertice, else an
    #    exception is thrown
    #
    #   @param vertex - The vertex for which visualizer is wanted
    #
    def get_visualizer(self, vertex):
        #  get the source and destination vertex elements
        #  and check to see if they exist
        v = self.vertices.get(vertex)
        try:
            if v is None:
                raise ValueError("Vertex " + vertex + " does not exist! First add the vertices to the graph.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return v.visualizer

    def get_data_structure_representation(self) -> dict:
        """
        Get the representation of the data structure as a dict
        Returns:
            dict: representing the JSON format before dumping to server
        """
        node_map = dict()  # map to reorder the nodes for building JSON
        nodes = []  # get the list nodes
        nodes_JSON = []  # array for list of nodes in json
        links_JSON = []  # array for building the links JSON - traverse the adj. lists

        # redirect for large graphs
        if (GraphAdjList.force_large_viz == True or (GraphAdjList.force_small_viz == False and
                                                     self.LargeGraphVertSize < len(self.vertices) and
                                                     self.are_all_vertices_located())):
            return self.get_data_structure_large_graph()

        # get the objects and add them to the array
        for elements in self.vertices.items():
            nodes.append(elements[1])

        # append all nodes representation to list of nodes
        for i in range(0, len(nodes)):
            node_map[nodes[i]] = i
            nodes_JSON.append(nodes[i].get_element_representation())

        # get all nodes in adj_list
        for a_list in self.adj_list.items():
            links_list = a_list[1]
            src_vert = self.vertices.get(a_list[0])
            #  get the source vertex index for the JSON (int)
            while links_list is not None:
                src_indx = node_map.get(src_vert)
                #  get the destination vertex index for the JSON (int)
                edge = links_list.value
                dest_vert = self.vertices.get(edge.tov)
                dest_indx = node_map.get(dest_vert)
                #  get link representation
                links_JSON.append((links_list.get_link_representation(
                                   edge._lvis,
                                   str(src_indx),
                                   str(dest_indx))))
                links_list = links_list.next
        json_str = {
            "nodes": nodes_JSON,
            "links": links_JSON
        }
        return json_str

    def get_data_structure_large_graph(self) -> dict:
        nodes = []
        node_map = dict()

        for element in self.vertices.items():
            nodes.append(element[1])

        nodes_json = []
        for k in range(0, len(nodes)):
            nodes_json.append([])
            node_map[nodes[k]] = k
            elvis = nodes[k].visualizer
            loc_str = []
            if elvis.location_x != float('inf') and elvis.location_y != float('inf'):
                loc_str.append(elvis.location_x)
                loc_str.append(elvis.location_y)
                nodes_json[k].append(loc_str)
            color = elvis.color
            nodes_json[k].append([color.red, color.green, color.blue, color.alpha])

        links_json = []
        for a_list in self.adj_list.items():
            list = a_list[1]
            src_vert = self.vertices.get(a_list[0])
            while list is not None:
                link_json=[]
                src_indx = node_map.get(src_vert)
                edge = list.value
                dest_vert = self.vertices.get(edge.tov)
                dest_indx = node_map.get(dest_vert)
                color = src_vert.get_link_visualizer(dest_vert).color
                link_json.append(src_indx)
                link_json.append(dest_indx)
                link_json.append([color.red, color.green, color.blue, color.alpha])
                links_json.append(link_json)
                list = list.next

        graph_alist_json = {
            "nodes": nodes_json,
            "links": links_json
        }

        return graph_alist_json




class Song:

    def __init__(self, artist: str = "", song: str = "", album: str = "", lyrics: str = "", release_date: str = ""):
        self._artist = artist
        self._song = song
        self._lyrics = lyrics
        self._album = album
        self._release_date = release_date

    @property
    def artist(self):
        return self._artist

    @artist.setter
    def artist(self, a):
        self._artist = a

    @property
    def song_title(self):
        return self._song

    @song_title.setter
    def song_title(self, s):
        self._song = s

    @property
    def album_title(self):
        return self._album

    @album_title.setter
    def album_title(self, a):
        self._album = a

    @property
    def lyrics(self):
        return self._lyrics

    @lyrics.setter
    def lyrics(self, l):
        self._lyrics = l

    @property
    def release_date(self):
        return self._release_date

    @release_date.setter
    def release_date(self, r):
        self._release_date = r




from bridges.dl_element import *
##
#  @brief This class can be used to instantiate Circular Doubly Linked
#  List Elements.
#
#	Structurally they are the same as doubly linked elements
#  except that each node constructed with the next and the previous pointers
#	points to itself.
#
#  User's implementation of the circularly linked list needs to ensure that
#  the last node's next pointer points to the first node and the first node's
#	previous pointer points to the last node, as the visualization generation
#	is dependent on this.
#
#  Elements have labels (string) that are displayed on the visualization.
#  Elements take an generic object E as a user defined parameter, which can be
#	any native type or object.
#
#  Elements contain a visualizer (ElementVisualizer) object for setting visual
#	attributes (color, shape, opacity, size), necessary for displaying them in a web
#	browser.
#
#	Elements also have a LinkVisualizer object that is used when they are
#	linked to another element, appropriate for setting link attributes, between
#	the element and its previous or next nodes.
#
#  	@author Kalpathi Subramanian, Mathhew McQuaigue
#
#	@date   7/17/16, 1/16/17 7/23/19
#
#   Circular doubly linked list tutorial: http://bridgesuncc.github.io/tutorials/CircularDoublyLinkedList.html
#

#
class CircDLelement(DLelement):

    def __init__(self, **kwargs) -> None:
        """
        Constructor for an Circularly Doubly Linked Element
        Args:
            label: THe label for this CircDLelement
            e: the generic element object that this CircDLelement will hold
            next: the next DLelement that should be assigned to the next pointer
            prev: THe previous DLelement that should be assigned to the next pointer
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'next' in kwargs and 'prev' in kwargs:
                if 'label' in kwargs:
                    super(CircDLelement, self).__init__(e=kwargs['e'], label=kwargs['label'],
                                                        next=kwargs['next'], prev=kwargs['prev'])
                else:
                    super(CircDLelement, self).__init__(e=kwargs['e'], next=kwargs['next'],
                                                        prev=kwargs['prev'])
            elif 'label' in kwargs:
                super(CircDLelement, self).__init__(e=kwargs['e'], label = kwargs['label'])
                self.next = self.prev = self
            else:
                super(CircDLelement, self).__init__(e=kwargs['e'])
                self.next = self.prev = self
        else:
            super(CircDLelement, self).__init__()
            self.next = self.prev = self

    def get_data_structure_type(self) -> str:
        """
        Gets the data structure type
        Returns:
            None
        """
        return "CircularDoublyLinkedList"

    @property
    def next(self):
        """
        Getter for the next element of this CircDLelement
        Returns:
            DLelement: the following element
        """
        return super(CircDLelement, self).next

    @next.setter
    def next(self, next) -> None:
        """
        Setter for the next element for this CircDLelement
        Args:
            (DLelement) next: the next element to be set
        Returns:
            None
        """
        DLelement.next.fset(self, next)

    @property
    def prev(self):
        """
        Getter for the prev element of this CircDLelement
        Returns:
            DLelement: the prev element in the list
        """
        return super(CircDLelement, self).prev

    @prev.setter
    def prev(self, prev) -> None:
        """
        Setter for the prev element of this CircDLelement
        Args:
            (DLelement) prev: The prev element to be set
        Returns:
            None
        """
        DLelement.prev.fset(self, prev)



from bridges.sl_element import *

##
# 	@brief This class can be used to instantiate Singly Linked
#	Circular List Elements.
#	Structurally they are the same as singly linked elements
#  except that each node constructed with the next point pointing to itself;
#  User's implementation of the circularly linked list needs to ensure that
#	the last node points to first node of the list, as the visualization generation is
#	dependent on this.
#
# 	Elements have labels (string) that are displayed on the visualization.
#  Elements take an generic object as a user defined parameter, E, which
#	can be any native type or object.
#
# 	Elements contains a visualizer (ElementVisualizer) object for setting visual
#	attributes (color, shape, opacity, size), necessary for displaying them in a web
#	browser.
#
#  Elements also have a LinkVisualizer object that is used when they are
#	linked to another element, appropriate for setting link attributes, between
#  an element and its next element.
#
#	@author Kalpathi Subramanian, Matthew McQuaigue
#
#	@date 6/22/16, 1/7/17, 5/17/17, 7/23/19
#
#   Circular singly linked list tutorial: http://bridgesuncc.github.io/tutorials/CircularSinglyLinkedList.html
#
class CircSLelement(SLelement):

    def __init__(self, **kwargs) -> None:
        """
        The constructor for a Circular Singly Linked Element
        Args:
            e: the generic object that this CircSLelement will hold
            label: The label of this CircSLelement
            next: The CircSLelement that should be assigned to the next pointer
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'label' in kwargs:
                if 'next' in kwargs:
                    super(CircSLelement, self).__init__(e=kwargs['e'], label=kwargs['label'],
                                                        next = kwargs['next'])
                else:
                    super(CircSLelement, self).__init__(e=kwargs['e'], label=kwargs['label'])
                    self.next = self
            else:
                super(CircSLelement, self).__init__(e=kwargs['e'])
                self.next = self
        else:
            super(CircSLelement, self).__init__()
            self.next = self

    def get_data_structure_type(self) -> str:
        """
        Gets the data structure type
        Returns:
            str: representing the data structure type
        """
        return "CircularSinglyLinkedList"

    @property
    def next(self):
        """
        Getter for the next element of this CircSLelement
        Returns:
            SLelement:  the element that follows this element
        """
        return super(CircSLelement, self).next

    @next.setter
    def next(self, n) -> None:
        """
        Setter for the next element in Circular list
        Args:
            n: the next element to be set
        Returns:
            None:
        """
        SLelement.next.fset(self, n)



import json
from bridges.symbol import *

#
# This class provides support for text labels to be associated with shapes. Labels have 
# width, height, font size.
# \sa Shape collection tutorial, http://bridgesuncc.github.io/tutorials/Symbol_Collection.html
#
#
# \sa Shape collection tutorial, http://bridgesuncc.github.io/tutorials/Symbol_Collection.html
#

class Label(Symbol):

    def __init__(self, label = None):
        super(Label, self).__init__()
        if label is not None:
            super(Label, self).label = label
        self._width = 100
        self._height = 50
        self._font_size = 12

    @property
    def width(self) -> int:
        """
        Getter for the width of label
        Returns:
            int: the width
        """
        return self._width

    @width.setter
    def width(self, w: int) -> None:
        """
        Setter for the width of label
        Args:
            (int) w: the width
        Returns:
            None
        """
        self._width = w

    @property
    def height(self) -> int:
        """
        Getter for the height of label
        Returns:
            int: the height
        """
        return self.height

    @height.setter
    def height(self, h: int) -> None:
        """
        Setter for the height of label
        Args:
            (int) h: the height
        Returns:
            None
        """
        self.height = h

    @property
    def font_size(self) -> int:
        """
        Getter for the font size
        Returns:
            int: representing the font size
        """
        return self._font_size

    @font_size.setter
    def font_size(self, size: int) -> None:
        """
        Setter for the font size
        Args:
            size: the size of the font
        Returns:
            None
        Raises:
            ValueError: if the size of the font is <= 0 or > 200
        """
        if size <= 0 or size > 200:
            raise ValueError("Please use font size between 0 and 200")
        else:
            self._font_size = size

    def get_dimensions(self):
        """
        Get the dimensions of the label
        Returns:
            list: representing the label dimensions
        """
        length = 0.09 * self.font_size * len(super(Label, self).label)
        x = self.get_location()[0]
        y = self.get_location()[1]

        return [x - length/2, x + length/2, y, y]

    def get_json_representation(self) -> dict:
        """
        Getter for the json representation
        Returns:
            dict: JSON represnted as a dict before the dump
        """
        ds_json = super(Label,self).get_json_representation()
        ds_json["name"] = super(Label, self).label
        ds_json["shape"] = "text"
        ds_json["font-size"] = self.font_size

        return ds_json



import socketio
import json


class SocketConnection:
    _sio = socketio.Client()
    _listeners = []

    def __init__(self, b):
        self.bridges = b

    @staticmethod
    def add_listener(to_add):
        print("Subscribing to keypress events..", to_add)
        SocketConnection._listeners.append(to_add)

    def setup_connection(self, user, assignment):
        try:
            url = "https://bridges-games.herokuapp.com"
            SocketConnection._sio.connect(url, transports=["websocket"])

            student_cred = {
                'user': user,
                'apikey': self.bridges.get_key(),
                'assignment': self.bridges.get_assignment_id()
            }
            student_cred = json.dumps(student_cred)

            print("passing student credentials to server..")
            SocketConnection._sio.emit('credentials', student_cred)

        except ConnectionError as e:
            print(e)

#    @_sio.on('announcement')
#    def announcement(*args):
#        print("announcement", *args)

    @_sio.on('keydown')
    def keydown(*args):
        #print(*args)
        for i in range(0, len(SocketConnection._listeners)):
            SocketConnection._listeners[i].key_press(*args)

    @_sio.on('keyup')
    def keyup(*args):
        for i in range(0, len(SocketConnection._listeners)):
            SocketConnection._listeners[i].key_press(*args)

    @_sio.on('disconnect')
    def disconnect(*args):
        print(*args, "disconnected")

    def send_data(self, dataframe):
        if SocketConnection._sio is not None:
            data = json.dumps(dataframe)
            SocketConnection._sio.emit('gamegrid:recv', data)

    def close(self):
        SocketConnection._sio.disconnect()
        SocketConnection._sio.wait()











class EleData:

    @property
    def cols(self):
        return self.ncols

    @cols.setter
    def cols(self, ncols):
        self.ncols = ncols

    @property
    def rows(self):
        return self.nrows

    @rows.setter
    def rows(self, nrows):
        self.nrows = nrows 

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, eledata):
        self._data = eledata

    @property
    def xll(self):
        return self._xll

    @xll.setter
    def xll(self, value):
        self._xll = value

    @property
    def yll(self):
        return self._yll

    @yll.setter
    def yll(self, value):
        self._yll = value

    @property
    def cellsize(self):
        return self._cellsize

    @cellsize.setter
    def cellsize(self, value):
        self._cellsize = value
    
    @property
    def maxVal(self):
        return self._maxVal

    @maxVal.setter
    def maxVal(self, value):
        self._maxVal = value



    def __init__(self):
        self.ncols = 0
        self.nrows = 0
        self._data = []
        self._xll = 0
        self._yll = 0
        self._cellsize = 0
        self.name = None
        self._maxVal = 0


from bridges.element_visualizer import *
from bridges.link_visualizer import *
import re
from decimal import Decimal
import traceback
import sys


##
# @brief This is the main superclass in BRIDGES for  deriving a number of
#  elements used  in building data structures, viz., arrays, lists, trees and graphs.
#  SLelement, DLelement, CircSLelement, CircDLelement, MLelement, TreeElement, BinTreeElement,
#  BSTElement, AVLTreeElement and KDTreeElement are all subclasses
#  (see class hierarchy above).  Element contains  two
#  visualizer objects (ElementVisualizer, LinkVisualizer) for specifying
#  visual attributes for nodes and links respectively. It also contains a label that
#  that can be displayed in BRIDGES visualizations.
#
#  All the tutorials under
#
#  http://bridgesuncc.github.io/tutorials/Overview.html
#
#  illustrate examples of using different types of Element objects and how to
#  manipulate their visual attributes.
#
#  @author Matthew Mcquaigue,  Kalpathi Subramanian
#
#  2017, 2018, 6/24/19
#
#
class Element:
    ids = 0

    def get_data_structure_type(self) -> str:
        """
        Get the data structure representation
        Returns:
            str: representing the data structure type
        """
        return "Element"

    def __init__(self, **kwargs) -> None:
        """
        Element constructor
        creates an Element Visualizer and unique identifier for the current element
        kwargs:
            (str) label: the string that is visible on the bridges Visualization
            (generic) val: value used to construct Element
            (Element) original: element object to copy
        Returns:
            None
        """
        Element.ids += 1
        self._link_visualizer = dict()
        self._ids = Element.ids
        self._identifier = str(self._ids)
        self._visualizer = ElementVisualizer()
        self._value = kwargs.get('val')
        self._label = kwargs.get('label') if kwargs.get('label') else "Default"

        if kwargs.get("color"):
            self.color = kwargs['color']

        if kwargs.get("opacity"):
            self.opacity = kwargs['opacity']

        if kwargs.get("original"):
            self._visualizer = ElementVisualizer(kwargs['original'].get_visualizer())
            self._label = str(kwargs['original'].label)
            self._value = kwargs['original']._value

    @property
    def value(self) -> object:
        """
        Getter for the value this element is holding
        Returns:
            generic object (application specific)
        """
        return self._value

    @value.setter
    def value(self, val) -> None:
        """
        Setter for the value of an element
        Args:
            val: the value (generic, application specific object)  this element will hold
        Returns:
            None
        """

        self._value = val

    @property
    def identifier(self) -> str:
        """
        Getter for the element identifier
        Return:
            str : element identifier (for internal use)
        """
        return self._identifier

    @identifier.setter
    def identifier(self, id: int) -> None:
        """
        Setter for the element identifier
        Args:
            int id: the identifier (for internal use)
        Returns:
            None
        """
        self._identifier = id

    @property
    def visualizer(self) -> ElementVisualizer:
        """
        Getter for the element visualizer
        Returns:
            ElementVisualizer 
        """
        return self._visualizer

    @visualizer.setter
    def visualizer(self, vis: ElementVisualizer) -> None:
        """
        Setter function for this element visualizer
        Args:
            vis: the element visualizer
        Returns:
            None
        """
        self._visualizer = vis

    def get_link_visualizer(self, el) -> LinkVisualizer:
        """
        Getter for  the link visualizer object that links this element to 
        another element specified by the argument.
        Args:
            el: the element terminating the link
        Returns:
            LinkVisualizer: of this element
        Raises:
            ValueError: if an element is not passed in
        """
        if el in self._link_visualizer:
            return self._link_visualizer[el]
        else:
            self._link_visualizer[el] = LinkVisualizer()
            return self._link_visualizer[el]

    def set_link_visualizer(self, el) -> None:
        """
        Setter for the link visualizer of this element
        Args:
            (Element) el: the terminating element of this link;
             creates a new link visualizer for this link
        Returns:
            None
        """
        self._link_visualizer[el] = LinkVisualizer()

    def remove_link_visualizer(self, el) -> None:
        """
        Deleter function for the lik visualizer of this element
        Args:
            (Element) el: the terminating element of the link; the link visualizer 
            for this link will be deleted
        Returns:
            None
        """
        self._link_visualizer.pop(el)

    @property
    def label(self):
        """
        Getter for the element's label
        Returns:
           str: the element's label 
        """
        return self._label

    @label.setter
    def label(self, label):
        """
        Setter for the element's label
        Args:
           str: the element's label 
        Returns:
           None
        """
        self._label = label

    @property
    def size(self):
        """
        Getter for the element's size
        Returns:
           int: element size (0-50)
        """
        return self.visualizer.size

    @size.setter
    def size(self, sz):
        """
        Setter for the element's size
        Args:
           sz: the element's size (0-50)
        Returns:
           None
        """
        self.visualizer.size = sz

    @property
    def color(self):
        """
        Getter for the element's color
        Returns:
           Color: element's color
        """
        return self.visualizer.color

    @color.setter
    def color(self, col):
        """
        Setter for the element's size
        Args:
           col: the element's color
        Returns:
           None
        """
        self.visualizer.color = col

    @property
    def opacity(self):
        """
        Getter for the element's opacity
        Returns:
           float: element size (0-1.0)
        """
        return self.visualizer.opacity

    @opacity.setter
    def opacity(self, op):
        """
        Setter for the element's opacity
        Args:
           op: the element's size (0-1.0)
        Returns:
           None
        """
        self.visualizer.opacity = op

    @property
    def shape(self):
        """
        Getter for the element's shape type
        Returns:
           element shape
        """
        return self.visualizer.shape

    @shape.setter
    def shape(self, shp):
        """
        Setter for the element's shape
        Args:
           shp: the element's size (0-1.0)
        Returns:
           None
        """
        self.visualizer.shape = shp

    @property
    def id(self) -> int:
        return self._ids

    def set_location(self, locX, locY):
        """
        Setter for the element's location
        Args:
           locX: the element's location in X
           locY: the element's location in Y
        Returns:
           None
        """
        self.visualizer.set_location(locX, locY)

    def get_locationX(self):
        """
        Getter for the element's location in X
        Returns:
           float: X coordinate of element's location
        """
        return self.visualizer.location_x

    def get_locationY(self):
        """
        Getter for the element's location in Y
        Returns:
           float: Y coordinate of element's location
        """
        return self.visualizer.location_y

    def get_element_representation(self):
        """
        Getter for the element's JSON representation (for internal use)
        Returns:
           str: JSON of element
        """
        json = {
            "name": str(self.label),
            "shape": str(self.visualizer.shape),
            "size": str(self.visualizer.size),
            "color": [str(self.visualizer.color.red), str(self.visualizer.color.green),
                      str(self.visualizer.color.blue), str(self.visualizer.color.alpha)]
        }
        loc_flag = not ((self.visualizer.location_x == Decimal('Infinity')) or (self.visualizer.location_y == Decimal('Infinity')))
        if loc_flag:
            json['location'] = [str(self.visualizer.location_x), str(self.visualizer.location_y)]
        return json

    def get_link_representation(self, lv, src, dest):
        """
        Getter for the JSON representation of  the element's link  (for internal use)
        Args:
           lv : link visualizer
           src: source vertex of link
           dest: destination vertex of link
        Returns:
           str: JSON of element's link from src to dest
        """
        link_json = lv.get_link_properties()
        link_json["source"] = src
        link_json["target"] = dest
        return link_json




##
#  @brief A class to hold actor movie data -- using IMDB dataset
#
#
#  @author Kalpathi Subramanian
#  @date 2/18/18
#
#

class ActorMovieIMDB():

    ##
	# This method initializes an actor movie object with an actor/movie pair.
	#
	# @param a - actor  (string)
	# @param m - movie  (string)
	# @param r - rating  (float)
    # @param genr - genres  (vector)
    def __init__(self, a = None, m = None, r = None, genr = None):
        if a is not None:
            self._actor = a
        else:
            self._actor = ""
        if m is not None:
            self._movie = m
        else:
            self._movie = ""
        if r is not None:
            self._rating = r
        else:
            self._rating = float
        if genr is not None:
            self._genres = genr
        else:
            self._genres = []

    ##
	#
	#  	 Get actor name
	#
	#  	 @return actor name (string)
	#
	#
    @property
    def actor(self):
        return self._actor

    @actor.setter
    def actor(self, a):
        self._actor = a

    @property
    def movie(self):
        return self._movie

    @movie.setter
    def movie(self, m):
        self._movie = m

    @property
    def rating(self):
        return self._rating

    @rating.setter
    def rating(self, r):
        self._rating = r

    @property
    def generes(self):
        return self._genres

    @generes.setter
    def generes(self, g):
        self._genres = g



#!/usr/bin/env python
from bridges.sl_element import *

##
# 	@brief This class can be used to instantiate Multi-list Elements.
#
#	This class extends SLelement (singly linked list element) to build multi-lists;
#	Multilist elements contain a tag (boolean) that indicates if the element 
#	contains a sublist or not; if the tag is true, then there is a sublist beginning
#	at this node and the starting point is the `sublist' field in the element. 
#	If the tag is false, then the list continues as a normal singly linked list. 
#	The sublists are re recursive: any sublist can have its own sublists and so on.
#	As in singly linked elements, the next pointer points to the following list element and 
#	each element contains a generic application specific object.
#
# 	Multi-list elements contain a visualizer (ElementVisualizer) object for setting
#	visual attributes (color, shape, opacity, size), necessary for displaying
#	them in a web browser.
#
#	Elements also have a LinkVisualizer object, that is used when they are linked to
#	another element, appropriate for setting link attributes, for instance, between
#	the current element and its next element. In this case, the link in question is that
#  	which connects the element to the following elements; a similar logic follows for
#	sublists.
#
# 	@author Kalpathi Subramanian, Matthew McQuaigue
#
#	@date 2018, 7/23/19
#
#	\sa There is a tutorial about Multi Lists : http://bridgesuncc.github.io/tutorials/ML.html
#
#
class MLelement(SLelement):

    def __init__(self, **kwargs) -> None:
        """
        Constructor for MLelement
        Args:
            (str) label: the label the SLelement will hold and show on bridges visualization
            (object) e: the generic object/value that this SLelement will hold
            (object) next: the next element that will be assigned to this SLelement next pointer
            sublist: the MLelement that is the beginning of a sublist
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'label' in kwargs:
                super(MLelement, self).__init__(e=kwargs['e'], label=kwargs['label'])
            else:
                super(MLelement, self).__init__(e=kwargs['e'])
        else:
            super(MLelement, self).__init__()
        if 'next' in kwargs:
            super(MLelement, self).next = kwargs['next']
        if 'sublist' in kwargs:
            self._sublist = kwargs['sublist']
            self._tag = True
            self.set_link_visualizer(kwargs['sublist'])
        else:
            self._sublist = None
            self._tag = False

    @property
    def sub_list(self):
        """
        Getter for the sublist at this node if exists
        Returns:
            Element: the sublist head element
        """
        return self._sub_list

    @sub_list.setter
    def sub_list(self, sl) -> None:
        """
        Setter for the start of a new sublist
        Args:
            sl: the MLelement that is the beginning of a sublist
        Returns:
            None
        """
        self._sub_list = sl
        if sl is not None:
            self._tag = True
            self.set_link_visualizer(sl)
        self._default_sub_list_nodes()

    def _default_sub_list_nodes(self) -> None:
        """
        function for setting sublist defaults
        Returns:
            None
        """
        #by default, color and shape sublist nodes to distinguish them  from
        #remaining nodes
        self.visualizer.color = "red"
        self.visualizer.shape = "square"

    def get_data_structure_type(self) -> str:
        """
        Getter for the data structure type
        Returns:
            str: representing the type
        """
        return "MultiList"

    @property
    def next(self):
        """
        Retrieves the element following this element
        Returns:
            MLelement
        """
        return super(MLelement, self).next

    @next.setter
    def next(self, n):
        SLelement.next.fset(self, n)

    @property
    def tag(self) -> bool:
        """
        Getter for the tag of the element
        Returns:
             bool
        """
        return self._tag

    @tag.setter
    def tag(self, t) -> None:
        """
        Setter for the tag of the element
        Args:
            t: boolean value
        Returns:
            None
        """
        self._tag = t

    def get_data_structure_representation(self) -> dict:
        """
        Getter for the data structure representation
        Returns:
            dict: representing the json structure before dumping
        """
        nodes = []
        nodes.clear()
        self.get_list_elements(nodes)
        nodes_JSON = []
        node_map = dict()
        for i in range(0, len(nodes)):
            node_map[nodes[i]] = i
            nodes_JSON.append(nodes[i].get_element_representation())
        links_JSON = []
        for j in range(0, len(nodes)):
            par = nodes[j]
            if par.tag:
                chld = par.sub_list
                #  sub list
                if chld is not None:
                    #  add the link
                    links_JSON.append(self.get_link_representation(par.get_link_visualizer(chld),
                                                                   str(node_map.get(par)), str(node_map.get(chld))))
            chld = par.next
            if chld is not None:
                #  add the link
                links_JSON.append(self.get_link_representation(par.get_link_visualizer(chld),
                                                               str(node_map.get(par)), str(node_map.get(chld))))
        json_dict = {
            "nodes": nodes_JSON,
            "links": links_JSON
        }
        return json_dict

    def get_list_elements(self, nodes):
        """
        Getter for the elements of the list
        Args:
            nodes: a list of the nodes
        Returns:
            element
        """
        self._get_list_elements_R(self, nodes)


    def _get_list_elements_R(self, list, nodes):
        el = list
        while el is not None:
            nodes.append(el)
            if el.tag:
                self._get_list_elements_R(el.sub_list, nodes)
            el = el.next



#!/usr/bin/env python
import base64
import wave
import json
import array
import math
from bridges.audio_channel import AudioChannel

##
#  @brief This is a class in BRIDGES for multi-channel audio data
#
#  This class contains a list of channels that contain sample data. The samples audio are 
#  encoded in 8, 16, 24, or 32 bit integers. All methods use 32 bits when taking and returning samples.
#
#  @author Luke Sloop
#
#  @date 2020, 1/31/2020
#
class AudioClip(object):
    def __init__(self, filepath: str="", sample_count: int=0, sample_rate: int=44100, num_channels: int=1, sample_bits: int=32) -> None:
        """
        AudioBase constructor
        Args:
            (int) sample_count: The total number of samples in this audio object
            (int) sample_rate: The number of samples in 1 second of audio default cd quality (44100)
        Returns:
            None
        """
        if filepath != "":
            self._from_filepath(filepath)
            return

        if sample_count > 1000000000:
            raise ValueError("Maximum frames exceeded with value %d" % self.get_sample_count())
        
        self.sample_count = sample_count
        self.sample_rate = sample_rate
        self.sample_bits = sample_bits

        self.num_channels = num_channels
        # Initialize the channels
        self._channels = []
        for i in range(self.num_channels):
            self._channels.append(AudioChannel(sample_count=self.sample_count, sample_bits=self.sample_bits)) 

    def _from_filepath(self, filepath: str):
        with wave.open(filepath, "r") as f:
            self.__init__(sample_count=f.getnframes(), sample_rate=f.getframerate(), num_channels=f.getnchannels(), sample_bits=f.getsampwidth()*8)

            self.framebytes = f.readframes(f.getnframes())
            framebytes = self.framebytes

            channel = 0
            count = 0
            for i in range(0, len(framebytes), self.get_sample_bytes()):
                if self.get_sample_bytes() != 3:
                    val = int.from_bytes(framebytes[i:i+self.get_sample_bytes()], byteorder='little', signed=True)
                    scaled = (val / ((2 ** self.get_sample_bits() / 2) - 1)) * ((2 ** 32 / 2) - 1)
                    self.set_sample(channel, count, val)
                else:
                    # Pad an empty byte to the start of 24 bit waves
                    val = int.from_bytes(b'\x00' + framebytes[i:i+self.get_sample_bytes()], byteorder='little', signed=True)
                    scaled = (val / ((2 ** self.get_sample_bits() / 2) - 1)) * ((2 ** 32 / 2) - 1)
                    self.set_sample(channel, count, val)
                
                channel += 1

                if channel >= f.getnchannels():
                    count += 1
                    channel = 0

    def get_num_channels(self) -> int:
        """
        Return the number of channels in this AudioClip. 1 for mono, 2 for stereo, etc.
        Returns:
            int: The number of channels of audio samples this object holds.
        """
        return self.num_channels

    def get_channel(self, index: int) -> AudioChannel:
        """
        Get the audio channel at index. The index should be less than get_num_channels().
        Args:
            (int) index: The index of the channel to get. 0 for front-left, 1 for front-right, etc.
        Returns:
            AudioChannel: The audio channel at index
        """
        return self._channels[index]

    def get_sample_rate(self) -> int:
        """
        Get the sample rate of this audio clip. This is the number of samples that are taken in one second.
        The default is CD Quality (44100).
        Returns:
            int: The sample rate or number of samples in 1 second of audio
        """
        return self.sample_rate

    def get_sample_count(self) -> int:
        """
        Get the number of samples in each channel of this audio object. 
        Each channel will contain this number of samples.
        Returns:
            int: The total number of samples in this audio object
        """
        return self.sample_count

    def get_sample(self, channel: int, index: int) -> int:
        """
        Get the sample at the index of the sample data from a specific channel.
        Args:
            (int) channel: The index of the channel to get. 0 for front-left, 1 for front-right, etc.
            (int) index: The index of the sample to get. From 0 - get_sample_count()
        Returns:
            int: The 32 bit sample
        """
        value = self.get_channel(channel).get_sample(index)

        minmax32 = (2 ** 32 // 2) - 1
        minmaxbit = (2 ** self.get_sample_bits() // 2) - 1

        scaled_sample = (value / minmaxbit) * minmax32
        return int(scaled_sample)

    def set_sample(self, channel: int, index: int, value: int) -> None:
        """
        Set the sample at the index of the sample data to value
        Args:
            (int) channel: The index of the channel to get. 0 for front-left, 1 for front-right, etc.
            (int) index: The index of sampledata to set which must be less than get_sample_count()
            (int) value: The value to set the sample to which must be a valid signed integer with bit length get_sample_bits()
        Returns:
            None
        """
        minmax32 = (2 ** 32 // 2) - 1
        if value < -minmax32 - 1:
            raise ValueError("Sample value out of minimum for signed 32 bit integer with value %d" % (value))
        if value > minmax32:
            raise ValueError("Sample value out of maxmium for signed 32 bit integer with value %d" % (value))
        
        minmaxbit = (2 ** self.get_sample_bits() // 2) - 1
        scaled_sample = (value / minmax32) * minmaxbit

        self.get_channel(channel).set_sample(index, int(value))

    def get_sample_bits(self) -> int:
        """
        Get the number of bits for the samples in this audio clip.  Will be 8, 16, 24, or 32 bits.
        Returns:
            int: The number of bits for each sample 
        """
        return self.sample_bits

    def get_sample_bytes(self) -> int:
        """
        Get the number of bytes for the samples in this audio clip.  Will be 1, 2, 3, or 4 bits.
        Returns:
            int: The number of bytes for each sample 
        """
        return self.sample_bits // 8

    def _get_type_code(self) -> str:
        if self.get_sample_bytes() == 1:
            return "b"
        elif self.get_sample_bytes() == 2:
            return "h"
        elif self.get_sample_bytes() == 3:
            return "f"
        elif self.get_sample_bytes() == 4:
            return "l"
        else:
            raise ValueError("Wave file sample bytes of unsupported length %d, supported lengths are 8, 16, 24, and 32 bit" % (self.get_sample_bytes() * 8))

    def get_data_structure_type(self) -> str:
        """
        Get the data structure type
        Returns:
            str : data structure type
        """
        return "Audio"

    def get_data_structure_representation(self) -> dict:
        """ Return a dictionary of the data in this audio file
        Returns:
            dict: The data of this audio file
        """
        json_dict = {}

        json_dict["encoding"] = "RAW"
        json_dict["numChannels"] = self.num_channels
        json_dict["sampleRate"] = self.get_sample_rate()
        json_dict["bitsPerSample"] = self.get_sample_bits()
        json_dict["numSamples"] = self.get_sample_count()

        # Combine all channel data
        framedata = []
        for i in range(self.sample_count):
            for c in range(self.num_channels):
                # Go straight to channel sample for correct bit data
                framedata.append(self._channels[c].get_sample(i))

        if self.get_sample_bytes() == 4:
            newarr = []
            for val in framedata:
                minmax32 = (2 ** 32 / 2.0) - 1
                minmax16 = (2 ** 16 / 2.0) - 1

                newval = (val / minmax32) * minmax16

                newarr.append(int(newval))

            json_dict["bitsPerSample"] = 16
            json_dict["samples"] = base64.b64encode(array.array("h", newarr).tobytes()).decode("utf-8")
        elif self.get_sample_bytes() != 3:
            json_dict["samples"] = base64.b64encode(array.array(self._get_type_code(), framedata).tobytes()).decode("utf-8")
        else:
            # 24 Bit samples have a pad that must be removed first
            shiftedbytes = bytearray()
            for sample in framedata:
                shiftedbytes += int.to_bytes(sample >> 8, length=3, byteorder='little', signed=True)

            json_dict["samples"] = base64.b64encode(shiftedbytes).decode("utf-8")

        return json_dict

    def display(self) -> None:
        """ Print information about this audio file to the console
        """
        print("Num Channels: %d, Sample Rate: %d, Sample Bits: %d, Num Samples: %d" % (self.num_channels, self.sample_rate, self.get_sample_bits(), self.get_sample_count()))

    def audio_from_json(json_dict: dict) -> 'AudioClip':
        """ Create an AudioClip from a json dictionary created by another AudioClip object
        Args:
            (dict) json_dict: The json dictionary created by another AudioClip object
        """
        audio = AudioClip(sample_count=json_dict["numSamples"], num_channels=json_dict["numChannels"], sample_rate=json_dict["sampleRate"], sample_bits=json_dict["bitsPerSample"])
        
        data = []
        if audio.get_sample_bytes() != 3:
            data = array.array(audio._get_type_code(), base64.b64decode(json_dict["samples"]))
        else:
            # No simple type to convert 24 bit base64 string to integer array
            chunk = base64.b64decode(json_dict["samples"])

            for i in range(0, len(chunk), audio.get_sample_bytes()):
                data.append(int.from_bytes(chunk[i:i+audio.get_sample_bytes()], byteorder='little', signed=True))

        for i in range(audio.get_sample_count() * audio.get_num_channels()):
            audio.set_sample(i % audio.get_num_channels(), int(i / audio.get_num_channels()), (data[i] / ((2 ** audio.get_sample_bits() / 2) - 1) * ((2 ** 32 / 2) - 1)))

        return audio


#!/usr/bin/env python
from bridges.sl_element import *

##
#
# 	@author Mihai Mehedint, Kalpathi Subramanian
#
#
# 	@brief This class is used to create doubly linked element objects.
#
#	This class extends Element and takes a generic parameter <E> representing
#	application specific data. This element forms the basic building block for
#	doubly linked lists. Doubly linked elements have two links,
# 	"next" and "previous", that point to the previous and succeeding nodes along the list.
#
#	Elements contain a visualizer (ElementVisualizer) object for setting visual
#	attributes (color, shape, opacity, size), necessary for displaying them in a web
#	browser.
#
#	Elements also have a LinkVisualizer object that is used when they are linked to
#	another element, appropriate for setting link attributes, such as in linked lists,
#	between the current element and its next or previous nodes.
#
#   @author Matthew Mcquaigue
#   @date  2018, 6/24/19
#
#   \sa Doubly Linked List tutorial : http://bridgesuncc.github.io/tutorials/DoublyLinkedList.html
#

#
class DLelement(SLelement):

    def __init__(self, *args, **kwargs) -> None:
        """
        Constructor for DLelement
        Kwargs:
            (object) e: the genereic application specific object that DLelement is holding
            next: the DLelement that should be assigned to the next pointer
            prev: the DLelement that should be assigned to the prev pointer
            (str) label: the label for this DLelement
        Return:
            None
        """
        super(DLelement, self).__init__(**kwargs)
        self._next = kwargs.get('next')
        self._prev = kwargs.get('prev')

    def get_data_structure_type(self) -> str:
        """
        This method gets the data structure type
        Returns:
            str: representing the data structure type
        """
        return "DoublyLinkedList"

    @property
    def next(self):
        """
        Getter for the next element in Dlelement
        Returns:
            element pointed to by this element
        """
        return self._next

    @next.setter
    def next(self, el):
        """
        Setter for the next element
        Args:
            el: element to be set 
        Returns:
            None
        """
        self._next = el

    @property
    def prev(self):
        """
        Getter for the prev element in Dlelement
        Returns:
            element that is prior to this element
        """
        return self._prev

    @prev.setter
    def prev(self, el):
        """
        Setter for the prev element
        Args:
            el: element to be set

        Returns:
            None
        """
        #remove any existing link visualizer from this node
        if self._prev is not None:
            self.remove_link_visualizer(self._prev)
        self._prev = el
        #  add a new link visualizer
        if el is not None:
            self.set_link_visualizer(el)

    def get_data_structure_representation(self) -> dict:
        """
        Getter for the json representation of this list
        Returns:
            dict: representing the JSON format
        """
        #  map to reorder the nodes for building JSON
        node_map = dict()
        #  get teh list nodes
        nodes = []
        SLelement.get_list_elements(self, nodes)
        #  generate the JSON of the list nodes
        nodes_JSON = []
        k = 0
        while k < len(nodes):
            node_map[nodes[k]] = k
            nodes_JSON.append(nodes[k].get_element_representation())
            k += 1
        links_JSON = []
        k = 0
        while k < len(nodes):
            par = nodes[k]
            nxt = par.next
            prv = par.prev
            if nxt is not None:
                #  add the link
                links_JSON.append(self.get_link_representation(par.get_link_visualizer(nxt), str(node_map.get(par)), str(node_map.get(nxt))))
            if prv is not None:
                #  add the link
                links_JSON.append(self.get_link_representation(par.get_link_visualizer(prv), str(node_map.get(par)), str(node_map.get(prv))))
            k += 1
        #  add the link
        json_dict = {
            "nodes": nodes_JSON,
            "links": links_JSON
        }
        return json_dict



#!/usr/bin/env python
## package: bridges.base
from bridges.color import *

# from Element import Element

##
#  @brief This class maintains the visual attributes of links that join
#  bridges elements.
#
#  Visual properties include color, thickness, and opacity.
#  Objects of this class are stored as part of the Element class.
#  Generally, a user will manipulate the LinkVisualizer returned from the
#  Element's getLinkVisualizer(Element it) method (which it is the bridges element
#	this element is linked to), and then set attributes using its methods. Links are
#  utilized in all types of linked lists, tree and graph structures.
#
#  Supported attribute values are as follows:<p>
#
#  <b>Supported Colors by name: CSS colors; see the Color class for full list, or  at
#	https://drafts.csswg.org/css-color-3/#svg-color</b>: <p>
#
#  <b> Color by RGBA Specification :</b>  Range: 0-255 for each component <p>
#
#  <b> Thickness: </b> Range : 0.0-50.0
#
#  <b> Opacity: </b> Range (0.0-1.0) </p>
#
#  @author Mihai Mehedint, Kalpathi Subramanian, Matthew McQuaigue
#
#  @date 2018, 6/24/19
#
#  \sa Example Tutorial at <br>
#  http://bridgesuncc.github.io/Hello_World_Tutorials/SLL.html
#
#


class LinkVisualizer():

    def __init__(self) -> None:
        """
        Constructor for the link visualizer'
        Retruns:
            None
        """
        self._color = Color(70, 130, 180, 1.0)
        self._thickness = 1.0
        self._label = ""

    @property
    def thickness(self) -> float:
        """
        Getter for the thickness of links visualization
        Returns:
            float: the thickness
        """
        return self._thickness

    @thickness.setter
    def thickness(self, th):
        """
        Setter for the thickness of links in visualization
        Args:
            (float) th: the thickness to be applied to the link
        Returns:
            None
        """
        self._thickness = th

    @property
    def color(self) -> Color:
        """
        Getter for the color of the link in the visualization
        Returns:
            Color
        """
        return self._color

    @color.setter
    def color(self, *args, **kwargs):
        """
        Setter for the color of the element in the bridges visualization
        Args:
            (optional) list: requires either 3 ints 0-255 for RGB and an optional 
            float 0.0-1.0 for alpha EX: color = [0, 255, 0, 1.0].
            (optional) str: string representing the element color. Supported web colors: 
            https://developer.mozilla.org/en-US/docs/Web/CSS/color_value
        Returns:
            None
        Raises:
            ValueError: if the color name provided is not available
        """
        self._color = Color(*args, *kwargs)

    @property
    def opacity(self):
        """
        Getter for the element opacity
        Returns:
            opacity : element opacity 
        """
        return self.color.alpha

    @opacity.setter
    def opacity(self, opacity):
        """
        Setter for the elementopacity
        Args:
            opacity : opacity value (0-1.0) to set
        Returns:
            None
        """
        self.color.alpha = opacity

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, l):
        self._label = l

    def get_link_properties(self):
        """
        Getter for the link properties
        Returns:
            dict: link properties
        """
        link_props = {
            "color": [str(self.color.red), str(self.color.green), str(self.color.blue), str(self.color.alpha)],
            "thickness": str(self.thickness),
        }
        if self.label is not "" or self.label is not None:
            link_props["label"] = self.label

        return link_props

    def get_label(self):
        """
        Getter for the link label
        Returns:
            string : link label
        """
        return self.label

    def set_label(self, label):
        """
        Setter for the element label
        Args:
            string : link label
        Returns:
            None
        """
        self.label = label





class Book:

    def __init__(self, authorName, authorBirth, authorDeath, title, lang, genre, subject, numChars, numWords,
                 numSentences, numDifficultWords, url, downloads):
        self._authorName = authorName
        self._authorBirth = authorBirth
        self._authorDeath = authorDeath
        self._title = title
        self._lang = lang
        self._genre = genre
        self._subject = subject
        self._url = url
        self._numChars = numChars
        self._numWords = numWords
        self._numSentences = numSentences
        self._numDifficultWords = numDifficultWords

    @property
    def name(self):
        return self._authorName

    @name.setter
    def name(self, n):
        self._authorName = n

    @property
    def birth(self):
        return self._authorBirth

    @birth.setter
    def birth(self, b):
        self._authorBirth = b

    @property
    def death(self):
        return self._authorDeath

    @death.setter
    def death(self, d):
        self._authorDeath = d

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, t):
        self._title = t

    @property
    def lang(self):
        return self._lang

    @lang.setter
    def lang(self, l):
        self._lang = l

    @property
    def genre(self):
        return self._genre

    @genre.setter
    def genre(self, g):
        self._genre = g

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, s):
        self._subject = s

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, u):
        self._url = u

    @property
    def num_chars(self):
        return self._numChars

    @num_chars.setter
    def num_chars(self, n):
        self._numChars = n

    @property
    def num_words(self):
        return self._numWords

    @num_words.setter
    def num_words(self, w):
        self._numWords = w

    @property
    def num_sentences(self):
        return self._numSentences

    @num_sentences.setter
    def num_sentences(self, n):
        self._numSentences = n

    @property
    def num_difficult_words(self):
        return self._numDifficultWords

    @num_difficult_words.setter
    def num_difficult_words(self, w):
        self._numDifficultWords = w




import json
import os
import pickle


class lru_cache():

    def __init__(self, max_cache_size: int = 0):
        self.max_cache_size = max_cache_size + 1
        self.lru = []
        if not os.path.isdir("./bridges_data_cache"):
            os.mkdir("./bridges_data_cache")


    def put(self, hash, content):
        try:
            with open("./bridges_data_cache/lru.txt", "rb") as fp:
                self.lru = pickle.load(fp)
        except:
            pass


        try: # Removes the location requested by the user from the LRU list
            self.lru.remove(hash)
        except:
            pass
        self.lru.insert(0, hash)

        # Removes least used map if there are more than 30 maps already saved
        if (len(self.lru) >= self.max_cache_size):
            re = self.lru[len(self.lru)-1]
            if (os.path.isfile(f"./bridges_data_cache/{re}")):
                os.remove(f"./bridges_data_cache/{re}")
            self.lru.remove(re)

        with open("./bridges_data_cache/lru.txt", "wb") as fp:   #Pickling
            pickle.dump(self.lru, fp)

        with open(f"./bridges_data_cache/{hash}", "wb") as f:   # write to file in cache
            pickle.dump(content, f)
        return



    def get(self, hash):
        with open("./bridges_data_cache/" + hash, "rb") as j:
            try:
                data = pickle.load(j)
            except:
                print("Error: Issue reading locally cached file")
        return data



    def inCache(self, file_name):
        if (os.path.isfile(f"./bridges_data_cache/{file_name}")):
            return True
        return False



#!/usr/bin/env python
from bridges.tree_element import *
##
# 	@brief This class is extended from the TreeElement class  and can be used to create
#	binary tree element objects.
#
# 	The BinTree element class is the building block for creating binary tree structures.
#  It contains two children (viz., left, right).
#
#  BinTreeElement contains a visualizer (ElementVisualizer) object for setting visual
#  attributes (color, shape, opacity, size), necessary for displaying them in a
#  web browser.
#
#  Elements also have a LinkVisualizer object, that is used when they are linked to
#  another element, appropriate for setting link attributes, for instance, between
#  the current element and its left or  right child
#
# @author Kalpathi Subramanian, Mihai Mehedint, Matthew McQuaigue
#
# @date   2018, 7/23/19
#
# \sa Binary tree tutorial, http://bridgesuncc.github.io/tutorials/BinTree.html
#

class BinTreeElement(TreeElement):

    def __init__(self, **kwargs) -> None:
        """
        Constructor for an empty Binary Tree Element
        kwargs:
            label: The label for the tree element that is displayed in the visualization
            e: the generic object that the binary tree element will hold
            left: the binary tree element assigned to child 0
            right: the binary tree element assigned to child 1
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'e' in kwargs and 'label' in kwargs:
                super(BinTreeElement, self).__init__(label=kwargs['label'], e=kwargs['e'])
            else:
                super(BinTreeElement, self).__init__(e=kwargs['e'])
        else:
            super(BinTreeElement, self).__init__()
        if 'left' in kwargs and 'right' in kwargs:
            super(BinTreeElement, self).add_child(kwargs['left'])
            super(BinTreeElement, self).add_child(kwargs['right'])
        else:
            super(BinTreeElement, self).add_child(None)
            super(BinTreeElement, self).add_child(None)


    def get_data_structure_type(self) -> str:
        """
        Get the data structure type
        Returns:
            str: representing the data structure type
        """
        return "BinaryTree"

    @property
    def left(self):
        """
        Getter for the left element for the binary tree
        Returns:
            TreeElement: left child of this element
        """
        return self.get_child(0)

    @left.setter
    def left(self, l) -> None:
        """
        Setter for the left element of a binary tree
        Args:
            l: the left element to set
        Returns:
            None
        """
        self.set_child(0, l)

    @property
    def right(self):
        """
        Getter for the right element for the binary tree
        Returns:
            TreeElement: the right child of this element
        """
        return self.get_child(1)

    @right.setter
    def right(self, r) -> None:
        """
        Setter for the right element of a binary tree
        Args:
            r: the right element to set
        Returns:
            None
        """
        self.set_child(1, r)



#!/usr/bin/env python
from bridges.element import *
import json

 ## @brief This class extends Element to represent general trees with
 #	arbitrary number of children.
 #
 # 	TreeElement nodes can have an arbitrary number of child nodes(held in
 # 	in a vector in the order in which they were added). The
 # 	visualization of trees assumes that the children are drawn in order
 # 	from left to right.
 #
 #  Tree Elements have labels (string) that are displayed on the visualization.
 #  Elements take an generic object E as a user defined parameter, which can be
 #  any native type or object.
 #
 #  Elements contain a visualizer (ElementVisualizer) object for setting visual
 #  attributes (color, shape, opacity, size), necessary for displaying them in a web
 #  browser.
 #
 #  Elements also have a LinkVisualizer object that is used when they are
 #  linked to another element, appropriate for setting link attributes, between parent
 # 	and child nodes.
 #
 # 	@author Matthew McQuaigue
 #
 # @date   2018, 7/23/19
 #  \sa Kd tree tutorial http://bridgesuncc.github.io/tutorials/Tree.html
 #
class TreeElement(Element):
    QUOTE = "\""
    COMMA = ","
    COLON = ":"
    OPEN_CURLY = "{"
    CLOSE_CURLY = "}"
    OPEN_PAREN = "("
    CLOSE_PAREN = ")"
    OPEN_BOX = "["
    CLOSE_BOX = "]"

    def __init__(self, **kwargs) -> None:
        """
        Constructor for Tree Element
        Kwargs:
            label: The label for the tree element that shows in visualization
            e: the generic object that the tree element will hold
            left: the tree element assigned to child 0
            right: the tree element assigned to child 1
        Returns:
            None
        """
        self.children = []
        if 'label' in kwargs and 'e' in kwargs:
            super(TreeElement, self).__init__(label = kwargs['label'], val = kwargs['e'])
        elif 'e' in kwargs:
            super(TreeElement, self).__init__(val = kwargs['e'])
        else:
            super(TreeElement, self).__init__()
        if 'left' in kwargs:
            self.children.append(kwargs['left'])
        if 'right' in kwargs:
            self.children.append(kwargs['right'])


    def get_data_structure_type(self) -> str:
        """
        Gets the data structure type
        Returns:
            str: representing the data structure type
        """
        return "Tree"

    def add_child(self, child) -> None:
        """
        Adds a child to this parent node
        Args:
            child: the child node to add
        Returns:
            None
        """
        self.children.append(child)

    def get_number_of_children(self) -> int:
        """
        Get the number of children at this node
        Returns:
            int: representing the number of children
        """
        return len(self.children)

    def set_child(self, index: int, child) -> None:
        """
        Adds a child to the node that will be added at the next open position
        Args:
            index: index to add child
            child: child to add to tree
        Returns:
            None
        Raises:
            ValueError: If the index is higher than the number of children
        """
        if index < len(self.children):
            self.children[index] = child
        else:
            raise ValueError("Index is higher than number of children")

    def get_child(self, index: int):
        """
        Gets a child at particular index
        Args:
            index: index to get child
        Returns:
            the child element at this index
        Raises:
            ValueError: if the index is higher than the number of children
        """
        if index < len(self.children):
            return self.children[index]
        else:
            raise ValueError("Index is higher than number of children")

    def get_data_structure_representation(self) -> dict:
        """
        Get the hierarchical JSON of the tree representation (internal use only)
        Returns:
            dict: representing the data structures json 
        """
        # json_dict = {
        #     "nodes": self._pre_order(self)
        #
        json_dict = ""
        json_dict += "{" + self.QUOTE + "nodes" + self.QUOTE + ":" + "{" + str(self._pre_order(self)) + "}" + "}"
        return json.loads(json_dict)

    def _pre_order(self, root) -> dict:
        """
        Use a preoreder traversal to directly extract a hierarchical
        JSON representation of the tree
        Args:
            root: the root of the tree structure
        Returns:
            dict: representing the json to be returned
        """
        json_str = ""
        children = ""
        link_props = ""
        elem_rep = ""
        t_str = str()
        num = root.get_number_of_children()
        if root is not None:
            elem_rep = root.get_element_representation()
            elem_rep = json.dumps(elem_rep)
            t_str = elem_rep[1:-1]
            json_str = t_str
            if root.get_number_of_children() > 0:
                json_str += "," + self.QUOTE + "children" + self.QUOTE + ":" + "["
            for k in range(0, root.get_number_of_children()):
                if root.get_child(k) is None:
                    json_str += "{" + self.QUOTE + "name" + self.QUOTE + ":" + \
                        self.QUOTE + "NULL" + self.QUOTE + "}" + ","
                else:
                    lv = root.get_link_visualizer(root.get_child(k))
                    json_str += "{"
                    if lv is not None:
                        json_str += self.QUOTE + "linkProperties" + self.QUOTE + ":" + "{" + \
                            self.QUOTE + "color" + self.QUOTE + ":" + \
                            "[" + \
                            str(lv.color.red) + "," + \
                            str(lv.color.green) + "," + \
                            str(lv.color.blue) + "," + \
                            str(lv.color.alpha) + "]" + "," + \
                            self.QUOTE + "thickness" + self.QUOTE + ":" + \
                            str(lv.thickness) + "}" + ","
                    else:
                        json_str += "linkProperties" + ":" + "{}" + ","

                    json_str += str(self._pre_order(root.get_child(k)))
                    json_str += "}" + ","
            if len(json_str) > 0:
                json_str = json_str[0:-1]
            json_str += "]"
        return json_str

        # k = 0
        # json_str = dict()
        # if root is not None:
        #     #  first get the node representationS
        #     elem_rep = root.get_element_representation()
        #     if root.get_number_of_children() > 0:
        #         elem_rep["children"] = []
        #     for i in range(0, root.get_number_of_children()):
        #         if root.get_child(i) is None:
        #             temp = {
        #                 "name": "NULL"
        #             }
        #             elem_rep["children"].append(temp)
        #         else:
        #             lv = root.get_link_visualizer(root.get_child(i))
        #             if lv is not None:
        #                 temp = {
        #                     "linkProperties": {
        #                         "color": [lv.color.red, lv.color.green, lv.color.blue, lv.color.alpha],
        #                         "thickness": lv.thickness
        #                     }
        #                 }
        #                 elem_rep["children"].append(temp)
        #             else:
        #                 temp = {
        #                     "linkProperties": {}
        #                 }
        #                 elem_rep["children"].append(temp)
        #             elem_rep['children'].append(self._pre_order(root.get_child(i)))
        #     json_str = elem_rep
        #
        # return json_str



from bridges.array import *


class Array2D(Array):
    """
    @brief This class can be used to create arrays of type Element<E>.

    @author 	Kalpathi Subramanian, Matthew McQuaigue

    @date  	10816, 51717, 53018

 	This class can be used to create arrays of type Element<E>  where E
 	is a generic object representing application specific data.

 	Arrays are internally represented as 1D arrays; currently 1D, 2D  and
 	3D arrays are supported.

    Example Tutorial at: http:bridgesuncc.github.iotutorialsArray.html (1D, 2D, and 3D Array)<br>
    """

    def __init__(self, rows=None, cols=None):
        """
        Create an array object with the specified dimensions
        Args:
            (int) rows: number of rows in array
            (int) cols: number of cols in array
        """
        super(Array2D, self).__init__()
        if rows is not None and cols is not None:
            dim = [cols, rows, 1]
            self.num_rows = rows
            self.num_cols = cols
            self.set_size(2, dim)
        else:
            self.size = 0
            self.num_rows = 0
            self.num_cols = 0



from bridges.link_visualizer import *

##
# @brief This class is used to represent the edges in a graph and will
# appear as links in the BRIDGES graph visualization.
# This object is used in graphs and graph algorithms such as DFS, BFS and shortest
# path algorithms that need to visit graph edges. The adjacency list
# representation uses them as the generic paramter, as SLelement<Edge>
# bridges represents Edges as links between pairs of elements
#
# @author Matthew McQuaigue
#
#

class Edge():

    def __init__(self, v1, v2, data=None) -> None:
        """
        Constructor for a edge
        Args:
            v1: first vertex of the edge
            v2: second vertex of the edge
            data: the data the edge will hold
        Returns:
            None
        """
        self._from_vertex = v1
        self._to_vertex = v2
        self._edge_data = data
        self._lvis = LinkVisualizer()

    @property
    def tov(self):
        """
        Getter for the to vertex
        Returns:
            vertex: terminating vertex of edge 
        """
        return self._to_vertex

    @property
    def fromv(self):
        """
        Getter for the from vertex
        Returns:
            vertex : source vertex  of edge
        """
        return self._from_vertex

    @property
    def destination(self):
        """
        Getter for the to vertex
        Returns:
            vertex: terminating vertex of edge
        """
        return self._to_vertex

    @property
    def source(self):
        """
        Getter for the from vertex
        Returns:
            vertex : source vertex  of edge
        """
        return self._from_vertex

    @property
    def thickness(self) -> float:
        """
        Getter for the link thickness
        Returns:
             float : link thickness (1.0-10.0 range)
        """
        return self._lvis.thickness

    @thickness.setter
    def thickness(self, th: float) -> None:
        """
        Setter for the thickness of edge
        Args:
            th: thickness to be applied (1.0-10.0 range)
        Returns:
            None
        """
        self._lvis.thickness = th

    @property
    def edge_data(self):
        """
        Getter for edge data
        Returns:
             str : data associated with this edge (generic object)
        """
        return self._edge_data

    @edge_data.setter
    def edge_data(self, data) -> None:
        """
        Setter for edge data
        Args:
            data: data for the edge to hold (generic object)
        Returns:
            None
        """
        self._edge_data = data

    @property
    def color(self):
        """
        Getter for edge color
        Returns:
                color of edge (see link visualizer class for setting options
        """
        return self._lvis.color

    @color.setter
    def color(self, color):
        """
        Setter for edge color
        Args:
            color : color to be set (see link visualizer class for setting options)
        Returns:
            None
        """
        self._lvis.color = color

    @property
    def opacity(self):
        """
        Getter for the element opacity
        Returns:
            opacity : element opacity
        """
        return self.color.alpha

    @opacity.setter
    def opacity(self, opacity):
        """
        Setter for the elementopacity
        Args:
            opacity : opacity value (0-1.0) to set
        Returns:
            None
        """
        self.color.alpha = opacity

    @property
    def label(self):
        return self._lvis.label

    @label.setter
    def label(self, l):
        self._lvis.label = l

    def get_edge(self):
        """
        Get this edge object
        Returns:
            self: this edge object
        """
        return self



import json

##
# @brief the SymbolCollection object is a container object to hold a collection of symbols
#  (rectangles, polygons, circles, labels).
#
# @author Matthew Mcquaigue
#
# @date 2018, 7/23/19
#
# \sa Symbol Collection tutorial, 
#     http://bridgesuncc.github.io/tutorials/Symbol_Collection.html
#
#
class SymbolCollection:

    def __init__(self):
        """
        Constructor for collection of symbols
        """
        self._symbols = dict()
        self._domainxmin = -100
        self._domainxmax = 100
        self._domainymin = -100
        self._domainymax = 100
        self._autoupdateviewport = True

    def setviewport(self, xmin, xmax, ymin, ymax):
        self._autoupdateviewport = False
        self._domainxmin = xmin
        self._domainxmax = xmax
        self._domainymin = ymin
        self._domainymax = ymax    
        
    def get_data_structure_type(self):
        """
        Getter for the data structure type
        Returns
            str: the data structure type
        """
        return "SymbolCollection"

    def add_symbol(self, s):
        """
        Add a symbol to the collection
        Args:
             s : symbol to be added
        Returns:
             none 
        """
        self._symbols[s.identifier] = s

    def update_axis_domains(self, s):
        """
        Update the bounding box of the symbol collection
        Args:
             s: input symbol to update dimensions
        Returns:
             none
        """
        dims = s.get_dimensions()

        if dims[0] < self._domainxmin:
            self._domainxmin = dims[0]
        if dims[1] > self._domainxmax:
            self._domainxmax = dims[1]

        if dims[2] < self._domainymin:
            self._domainymin = dims[2]
        if dims[3] > self._domainymax:
            self._domainymax = dims[3]

    def get_data_structure_representation(self):
        """
        Getter for the data structure's JSON representatoin
        Returns:
            str: the data structure representation
        """
        symbol_json = []
        for key, value in self._symbols.items():
            if (self._autoupdateviewport):
                self.update_axis_domains(value)
            symbol_json.append(value.get_json_representation())

        final_json = {
            "domainX": [self._domainxmin, self._domainxmax],
            "domainY": [self._domainymin, self._domainymax],
            "symbols": symbol_json
        }
        return final_json




class Shakespeare:

    def __init__(self, title: str = "", type: str = "", text: str = ""):
        self._title = title
        self._type = type
        self._text = text

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, t):
        self._title = t

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, t):
        self._type = t

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, t):
        self._text = t



from bridges.bin_tree_element import *
##
#    @brief The BSTElement class is the building block for creating binary search trees.
#
#   The BSTElement class is the building block for creating binary search tree structures.
#    It contains two children (viz., left, right), and a search key, to be used
#    in search operations .
#
#    BSTElement contains a visualizer (ElementVisualizer) object for setting visual
#    attributes (color, shape, opacity, size), necessary for displaying them in a
#    web browser.
#
#    BST Elements also have a LinkVisualizer object, that is used when they are linked to
#    another element, appropriate for setting link attributes, for instance, between
#    the current element and its left or right child
#
#  	@author Kalpathi Subramanian, Mihai Mehedint, Matthew McQuaigue
#
#  	@date 6/22/16, 1/7/17, 5/17/17, 7/23/19
#
#  	@brief This class extends the BinTreeElement class by adding a 'key' value
#  	for use in a binary search tree implementations.
#
#	Binary Search Tree tutorial, http://bridgesuncc.github.io/tutorials/BinarySearchTree.html
#


class BSTElement(BinTreeElement):

    def __init__(self, **kwargs) -> None:
        """
        Constructor bst element
        Kwargs:
            (str) key: The label for the tree element that shows in visualization
            (generic) e: the generic object that the tree element will hold
            (BinTreeElement) left: the tree element assigned to child 0
            (BinTreeElement) right: the tree element assigned to child 1
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'left' in kwargs and 'right' in kwargs:
                super(BSTElement, self).__init__(e=kwargs['e'], left = kwargs['left'], right = kwargs['right'])
            else:
                super(BSTElement, self).__init__(e = kwargs['e'])
        else:
            super(BSTElement, self).__init__()
        if 'key' in kwargs:
            self._key = kwargs['key']

    def get_data_structure_type(self) -> str:
        """
        Get the data structure representation
        Returns:
            str
        """
        return "BinarySearchTree"

    @property
    def key(self) -> str:
        """
        Getter for the bst element key
        Returns:
            str: the key label
        """
        return self._key

    @key.setter
    def key(self, k: str) -> None:
        """
        Setter for the bst element key
        Args:
            key: the key for the element
        Returns:
            None
        """
        self._key = k

    @property
    def left(self):
        """
        Getter for the left element in BST
        Returns:
            BSTElement: the left child of this element
        """
        return super(BSTElement, self).left

    @left.setter
    def left(self, l):
        """
        Setter for the left element in BST
        Args:
            l: the value for the left child to be set as
        """
        self.set_child(0, l)

    @property
    def right(self):
        """
        Getter for the right child in BST
        Returns:
            BSTElement: the right child of the element
        """
        return super(BSTElement, self).right

    @right.setter
    def right(self, r):
        """
        Setter for the right element in BST
        Args:
            r: the value for the right child to be set as
        """
        self.set_child(1, r)

    def get_element_representation(self) -> dict:
        """
        Augment the element with the "key" field
        Returns:
            dict: representing the json of this tree
        """
        orig_json_str = dict(super(BSTElement, self).get_element_representation())
        key_json = {
            'key': self.key
        }
        orig_json_str.update(key_json)
        return orig_json_str



import json
from bridges.symbol import *

##
# @brief This class defines a polygon and is part of the symbol collection.
#
# A polygon has a set of vertices, with vertices connected by line
# segments. It differs from the polyline in the sense that the last
# and first vertex are connect to close the shape.
#
# @author David Burlinson, Kalpathi Subramanian
#
# @date 2018, 7/23/19
#
# \sa Shape collection tutorial, http://bridgesuncc.github.io/tutorials/Symbol_Collection.html
#
class Polyline(Symbol):

    def __init__(self, pts = None):
        super(Polyline, self).__init__()
        if pts is not None:
            self._points = pts
        else:
            self._points = []
        self._shape_type = "polyline"

    def get_name(self):
        return "polyline"

    def add_point(self, x, y):
        fx = float(x)
        fy = float(y)

        if(fx > float('-inf') and fx < float('inf') and fy > float('-inf') and fy < float('inf')):
            self._points.append(fx)
            self._points.append(fy)

    @property
    def points(self):
        return self._points

    @points.setter
    def points(self, pts):
        self._points = pts

    def get_dimensions(self):
        minx = float('inf')
        miny = float('inf')
        maxx = float('-inf')
        maxy = float('-inf')

        for i in range(0, len(self.points), 2):
            x = self.points[i]
            y = self.points[i+1]
            if (x < minx):
                minx = x
            if (x > maxx):
                maxx = x
            if (y < miny):
                miny = y
            if (y > maxy):
                maxy = y

        return [minx, maxx, miny, maxy]

    def translate(self, tx, ty):
        for k in range(0, len(self.points), 2):
            self.points[k] = self.points[k] + tx
            self.points[k + 1] = self.points[k+1] + ty

    def rotate(self, angle):
        center = []
        center = self.get_center(center)
        transl = []
        transl[0] = -center[0]
        transl[1] = -center[1]
        self.translate(transl[0], transl[1])
        for k in range(0, len(self.points), 2):
            tmp = []
            tmp[0] = self.points[k]
            tmp[1] = self.points[k+1]
            tmp = self.rotate_point(tmp, angle)
            self.points[k] = tmp[0]
            self.points[k+1] = tmp[1]
        transl[0] = center[0]
        transl[1] = center[1]
        self.translate(transl[0], transl[1])

    def scale(self, sx, sy):
        center = []
        center = self.get_center(center)
        transl = []
        transl[0] = -center[0]
        transl[1] = -center[1]
        self.translate(transl[0], transl[1])
        for k in range(0, len(self.points), 2):
            self.points[k] = self.points[k] * sx
            self.points[k+1] = self.points[k+1] * sy
        transl[0] = center[0]
        transl[1] = center[1]
        self.translate(transl[0], transl[1])

    def get_center(self, center):
        bbox = []
        bbox[0] = bbox[1] = 100000.0
        bbox[2] = bbox[3] = -10000.0
        for k in range(0, len(self.points), 2):
            if self.points[k] < bbox[0]:
                bbox[0] = self.points[k]
            if self.points[k] > bbox[2]:
                bbox[2] = self.points[k]
            if self.points[k+1] < bbox[1]:
                bbox[1] = self.points[k+1]
            if self.points[k+1] > bbox[3]:
                bbox[3] = self.points[k+1]
        center[0] = bbox[0] + (bbox[2] - bbox[0]) / 2.0
        center[1] = bbox[1] + (bbox[3] - bbox[1]) / 2.0
        return center

    def get_json_representation(self):

        ds_json = super(Polyline, self).get_json_representation()
        ds_json["name"] = self.label
        ds_json["shape"] = self.shape_type
        ds_json["points"] = self.points

        return ds_json



##
#  @brief This is a class in BRIDGES a channel of audio data
#
#  This class contains one channel of 8, 16, 24, or 32 bit audio samples.
#
#  @author Luke Sloop
#
#  @date 2020, 1/31/2020
#
class AudioChannel(object):
    def __init__(self, sample_count: int=0, sample_bits: int=32) -> None:
        """
            AudioChannel constructor
            args:
                (int) sample_count: The total number of samples in this audio channel
            Returns:
                None
        """
        self.__data = [0 for _ in range(sample_count)]
        self._sample_count = sample_count
        self._sample_bits = 32

    def set_sample(self, index: int, value: int) -> None:
        """
            Set the sample at index. The sample value should be a valid signed integer with get_sample_bits() number of bits.
            args:
                (int) index: The index on this channel to set
                (int) value: The signed integer value to set the sample to
            Returns:
                None
        """
        minmax = (2 ** self.get_sample_bits() // 2) - 1
        if value < -minmax - 1:
            raise ValueError("Sample value out of minimum for signed %d bit integer with value %d" % (self.get_sample_bits(), value))
        if value > minmax:
            raise ValueError("Sample value out of maxmium for signed %d bit integer with value %d" % (self.get_sample_bits(), value))

        self.__data[index] = value

    def get_sample(self, index: int) -> int:
        """
            Get the sample at index. The sample will be a signed integer with get_sample_bits() number of bits.
            args:
                (int) index: The sample index on this channel to get
            Returns:
                int: The sample at index that is an integer within the range of an integer of get_sample_bits() number of bits.
        """
        return self.__data[index]

    def get_sample_count(self) -> int:
        """
            Get the number of samples or length of this channel.
            Returns:
                int: The number of samples in this channel.
        """
        return self._sample_count

    def get_sample_bits(self) -> int:
        """
            Get the number of bits that make the samples in this channel.
            Will be 8, 16, 24, or 32.
            Returns:
                int: The number of bits that make the samples in this channel.
        """
        return self._sample_bits


from bridges.element import *
import traceback
##
#
#  @brief The GraphAdjMatrix class can be used to represent adjacency matrix based
#      graphs in BRIDGES
#
#  The GraphAdjMatrix class can be used to represent adjacency matrix based  graphs
#  in BRIDGES; it takes 2 generic parameters: (1) K, which is an orderable
#  key value used in accessing vertices (in constant time) using a hashmap. This
#  permits data sets that need to be accessed by keys that are strings, and
#  (2) E, an application defined type, and used in the Edge representation.
#  The class is simply a wrapper  around the Java Hashmap class
#  and, thus, derives all its operations from it.
#  BRIDGES provides methods to visualize the graph  and its contents.
#
#  The vertices of the graph are held in a Java hashmap, for near constant time access;
#  this lets us use strings or integral ids for vertices. The edges are accessed
#    by a second hashmap from each vertex, again assuring near constant access time.
#  Each edge contains the terminating vertex id and weight, as defined by  the Edge
#    class structure.
#
#  Convenience methods are provided to add vertices and edges to the graph. Edges
#  are retrieved by using the dual hashmap, given the vertex ids of the edge.
#
#  @author Kalpathi Subramanian, Mihai Mehedint
#
#
#  \sa Example tutorial at http://bridgesuncc.github.io/tutorials/Graph_AM.html
#
#
class GraphAdjMatrix():
    ##
    #     Constructor
    #
    def __init__(self):
        self._vertices = dict()
        self._matrix = dict()
        self._edge_data = dict()

    ##
    #
    #    This method gets the data structure type
    #
    #    @return  The date structure type as a string
    #
    #
    def get_data_structure_type(self):
        return "GraphAdjacencyMatrix"

    ##
    #     Adds a new vertex to the graph, initializes the  adjacency
    #     list; user is responsible for checking if the vertex already
    #     exists. This method will replace the value for this key
    #
    #     @param k - vertex key value
    #     @param e - user specified data, part of the vertex data
    #
    #
    def add_vertex(self, k, e):
        #  note: it is the user's responsibility to  check
        #  for duplicate vertices
        self.vertices[k] = Element(val = e)
        self.vertices.get(k).label = str(k)

        #  create a hashmap for this vertex
        self._matrix[k] =  dict()
        #  fill up this vertex's row and column elements
        for k, element in self.vertices.items():
            (self._matrix.get(k))[element.label] = 0 #  row
            (self._matrix.get(element.label))[k] =  0 #  col

    ##
    #    Adds a new edge to the graph, adds it to the index corresponding to
    #    the source, destination vertex ids;  this version of the method assumes
    #    an edge weight of 1 (unweighted graph); user is responsible for checking if the
    #    vertices already exist, else an exception is thrown.
    #
    #    @param src - source vertex of edge
    #    @param dest - destination  vertex of edge
    #
    #
    def add_edge(self, src, dest, weight = None):
        #  check to see if the two vertices exist, else
        #  throw an exception
        try:
            if self.vertices.get(src) is None or self.vertices.get(dest) is None:
                raise ValueError("Vertex " + src + " or " + dest + " does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        if weight is not None:
            (self._matrix.get(src))[dest] = weight
        else:
            (self._matrix.get(src))[dest] =  1

    @property
    def vertices(self):
        return self._vertices

    def set_vertex_data(self, src, vertex_data):
        try:
            if self.vertices.get(src) is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self.vertices.get(src).value = vertex_data

    def get_vertex_data(self, src):
        try:
            if self.vertices.get(src) is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return self.vertices.get(src).value

    def set_edge_data(self, src, dest, data):
        try:
            if self.vertices.get(src) is None or self.vertices.get(dest) is None:
                raise ValueError("Vertex " + src + " or " + dest +
					" does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self._edge_data.get(src)[dest] = data

    def get_edge_data(self, src, dest):
        try:
            if self.vertices.get(src) is None or self.vertices.get(dest) is None:
                raise ValueError("Vertex " + src + " or " + dest +
                                 " does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return self._edge_data.get(src)[dest]

    ##
    #
    #    Gets the adjacency matrix
    #
    #    @return - the graph's adjacency matrix
    #
    def get_adjacency_matrix(self, key = None):
        if key is None:
            return self._matrix
        else:
            return self._matrix.get(key)

    ##
    #
    #     This is a convenience method to simplify access to the link visualizer;
    #     the method assumes the vertex names point to existing vertices, else an exception
    #     is thrown
    #
    #
    def get_link_visualizer(self, src, dest):
        #  get the source and destination vertex elements
        #  and check to see if they exist
        v1 = self.vertices.get(src)
        v2 = self.vertices.get(dest)
        try:
            if v1 is None or v2 is None:
                raise ValueError("Vertex " + src + " or " + dest + " does not exist! First add the vertices to the graph.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return v1.get_link_visualizer(v2)

    ##
    #
    #    This is a convenience method to simplify access to the element visualizer;
    #    the method assumes the vertex name points to an existing vertice, else an
    #    exception is thrown
    #
    #
    def get_visualizer(self, vertex):
        #  get the source and destination vertex elements
        #  and check to see if they exist
        v = self.vertices.get(vertex)
        try:
            if v is None:
                raise ValueError("Vertex " + vertex + " does not exist! First add the vertices to the graph.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return v.visualizer

    ##
    #     Get the JSON representation of the the data structure
    #
    def get_data_structure_representation(self):
        node_map = dict()
        nodes = []
        nodes_json = []
        for key, value in self.vertices.items():
            nodes.append(value)

        for k in range(0, len(nodes)):
            node_map[nodes[k]] = k
            nodes_json.append(nodes[k].get_element_representation())

        links_json = []
        for el_src in self._matrix.items():
            src_vert = self.vertices.get(el_src[0])
            src_indx = node_map.get(src_vert)

            for el_dest in el_src[1].items():
                dest_vert = self.vertices.get(el_dest[0])
                if el_dest[1] > 0:
                    dest_indx = node_map.get(dest_vert)
                    links_json.append(src_vert.get_link_representation(src_vert.get_link_visualizer(dest_vert),
                                                                       str(src_indx), str(dest_indx)))
        json_str = {
            "nodes": nodes_json,
            "links": links_json
        }

        return json_str



import json
from bridges.symbol import *

##
# @brief This class defines a circle and is part of the symbol collection.
# A circle  has a radius and a center, which is also its location
#
# @author Matthew Mcquaigue
# @date 2018, 7/23/19
#
# Shape collection tutorial, http://bridgesuncc.github.io/tutorials/Symbol_Collection.html
#
#
class Circle(Symbol):

    def __init__(self, **kwargs) -> None:
        """
        Constructor for a circle symbol
        Args:
            (int) locx: x location of the circle
            (int) locy: y location of the circle
            (float) r: radius of the circle
        Returns:
            None
        Raises:
            ValueError: if the radius is not a positive number
        """
        super(Circle, self).__init__()
        if 'locx' in kwargs and 'locy' in kwargs:
            self.set_location(kwargs['locx'], kwargs['locy'])
        else:
            self.set_location(0.0, 0.0)
        if 'r' in kwargs:
            if kwargs['r'] < 0:
                raise ValueError("Radius value needs to be positive")
            self.radius = kwargs['r']
        else:
            self.radius = 10
        self.shape_type = "circle"

    def get_name(self) -> str:
        """
        Gets the name of the shape/symbol
        Returns:
            str: representing the name
        """
        return "circle"

    @property
    def radius(self) -> float:
        """
        Getter function for the radius of this circle
        Returns:
            float: representing radius
        """
        return self._radius

    @radius.setter
    def radius(self, r) -> None:
        """
        Setter for the radius of this circle
        Args:
            (float) r: radius to be set
        Returns:
            None
        Raises:
            ValueError: if the radius is not a positive number
        """
        if r < 0:
            raise ValueError("Illegal value for radius. Must be positive")
        self._radius = r

    def set_circle(self, locx, locy, r) -> None:
        """
        Set the location and size of the circle
        Args:
            (int) locx: x coordinate of the center of the circle
            (int) locy: y coordinate of the center of the circle
            (float) r: radius of the circle
        Returns:
            None
        Raises:
            ValueError: if the radius is not a positive number
        """
        self.set_location(locx, locy)
        if r < 0:
            raise ValueError("Radius value needs to be positive")
        self.radius = r

    def get_dimensions(self) -> list:
        """
        Getter for the dimensions for the circle
        Returns:
            list: bounding box of the circle (xmin, xmax, ymin, ymax)
        """
        dims = []
        location = self.get_location()

        dims.append(location[0] - self.radius)
        dims.append(location[0] + self.radius)
        dims.append(location[1] - self.radius)
        dims.append(location[1] + self.radius)

        return dims

    def translate(self, tx:float, ty:float):
        """
        translate the circle
        Arrgs:
            (float) tx: x translation
            (float) ty: y translation
        """
        center = self.get_location()
        center = self.translate_point(center, tx, ty)
        self.set_location(center[0], center[1])

    def scale(self, scale:float):
        """
        Scale the circle. Only the radius needs to be scaled, using a single scale value
        Args:
            (float) scale: scale factor
        """
        # scale onlt the radius
        # center stays the same
        self.radius *= scale

    def get_json_representation(self) -> dict:
        """
        Get the json representation of the Circle object
        Returns:
            dict: representing the JSON 
        """
        ds_json = super(Circle, self).get_json_representation()
        ds_json["name"] = self.label
        ds_json["shape"] = self.shape_type
        ds_json["r"] = self.radius

        return ds_json



from bridges.color import *
import math
##
# @brief This is an abstract class for deriving a
#  number of Symbol shape objects, for use in a SymbolCollection.
#  Symbols correspond to a simplified subset of SVG paths
#  and shapes for custom visual representations in BRIDGES.
#
#  Users will typically one of the subclasses of this object for creating shapes
#
#  Currently shapes supported are rectangle, circle, polygon, polyline and label; each shape
#  has a name, location (x, y) and appropriate geometric and non-geometric attributes
#
# @author David Burlinson, Kalpathi Subramanian
# @date 12/24/18, 7/12/19
#



class Symbol:
    _ids = 0 #id for this symbol. Class level so each object has unique id

    def __init__(self):
        """
        Constructor for a Symbol
        """
        self._identifier = str(Symbol._ids)
        self._label = ""
        self._fill_color = Color("white")
        self._stroke_color = Color("white")
        self._opacity = 1.0
        self._stroke_width = 1.0
        self._stroke_dash = 1
        self._location_y = 0.0
        self._location_x = 0.0
        self._shape_type = "circle"
        Symbol._ids += 1

    @property
    def label(self) -> str:
        """
        Getter for symbol label
        Returns:
            str : the label of this shape
        """
        return self._label

    @label.setter
    def label(self, label: str) -> None:
        """
        Setter for symbol label
        Args:
            label: to be set
        Returns:
            str
        """
        self._label = label

    @property
    def identifier(self) -> str:
        """
        Getter for the symbols identifier
        Returns:
            str
        """
        return self._identifier

    @property
    def fill_color(self) -> Color:
        """
        Getter for the fill color
        Returns:
            Color : the current fill color of this symbol
        """
        return self._fill_color

    @fill_color.setter
    def fill_color(self, *args, **kwargs) -> None:
        """
        Setter for the fill color
        Args:
            color: the color to set for the fill
        Returns:
            None
        """
        self._fill_color = Color(*args, **kwargs)

    @property
    def stroke_color(self) -> Color:
        """
        Getter for the stroke color
        Returns:
            Color : the stroke (boundary) color of the shape
        """
        return self._stroke_color

    @stroke_color.setter
    def stroke_color(self, *args, **kwargs) -> None:
        """
        Setter for the stroke color
        Args:
            color: the stroke (boundary) color to set
        Returns:
            None
        """
        self._stroke_color = Color(*args, **kwargs)

    @property
    def stroke_width(self) -> float:
        """
        Getter for the stroke width
        Returns:
            float : the stroke width of the shape
        """
        return self._stroke_width

    @stroke_width.setter
    def stroke_width(self, width: float) -> None:
        """
        Setter for the stroke width
        Args:
            width: the stroke width to set
        Returns:
            None
        Raises:
            value error
        """
        if width <= 0.0 or width > 10.0:
            raise ValueError("Stroke width must be between 0 and 10")
        else:
            self._stroke_width = width

    @property
    def opacity(self) -> float:
        """
        Getter for opacity
        Returns:
             float : opacity of the shape
        """
        return self._opacity

    @opacity.setter
    def opacity(self, o: float):
        """
        Setter for opacity
        Args:
            o: the opacity value to set (must be in 0-1.0 range) 
        Returns:
            None
        Raises: value error
        """
        if o <= 0.0 or o > 1.0:
            raise ValueError("Opacity must be between 0.0 and 1.0")
        else:
            self._opacity = o

    @property
    def stroke_dash(self) -> float:
        """
        Getter for stroke_dash
        Returns:
             float : the stroke texture
        """
        return self._stroke_dash

    @stroke_dash.setter
    def stroke_dash(self, dash: float):
        """
        Setter for stroke_dash
        Args:
            dash: the stroke_dash value to set (0-10 range)
        Returns:
            None
        Raises: value error
        """
        if dash < 0 or dash > 10:
            raise ValueError("Dash must be between 0 and 10 (inclusive)")
        else:
            self._stroke_dash = dash

    @property
    def shape_type(self):
        return self._shape_type

    @shape_type.setter
    def shape_type(self, shape):
        self._shape_type = shape

    def set_location(self, x: float, y: float) -> None:
        """
        Setter for the location of the  center of the symbol
        Args:
            x: x location 
            y: y location
        Returns:
            None
        Raises:
            Value error
        """
        if x > float('-inf') and x < float('inf') and y > float('-inf') and y < float('inf'):
            self._location_x = x
            self._location_y = y
        else:
            raise ValueError("Coordinates must be real numbers")

    def get_location(self) -> list:
        """
        Getter for the location of a symbol
        Returns:
            list : the x and y coordinates of the shape's current location
        """
        return [self._location_x, self._location_y]

    def get_dimensions(self) -> list:
        """
        Getter for the dimensions
        Returns:
            list : the bounding box of this shape (xmin, xmax, ymin, ymax)
        """
        return [0.0, 0.0, 0.0, 0.0]

    def translate_point(self, pt, tx, ty):
        pt[0] += tx
        pt[1] += ty
        return pt

    def scale_point(self, pt, sx, sy):
        pt[0] *= sx
        pt[1] *= sy
        return pt

    def rotate_point(self, pt, angle):
        angle_r = math.radians(angle)
        c = math.cos(angle_r)
        s = math.sin(angle_r)

        tmp = [pt[0] * c - pt[1] * s, pt[0] * s + pt[1] * c]
        pt[0] = float(tmp[0])
        pt[1] - float(tmp[1])
        return pt

    def get_json_representation(self) -> dict:
        """
        Get the json representation of the Symbol class
        Returns:
            dict : the JSON representation
        """
        ds = {
            # "fill": [self.fill_color.red, self.fill_color.green, self.fill_color.blue, self.fill_color.alpha],
            # "opacity": self.opacity,
            # "stroke": [self.stroke_color.red, self.stroke_color.green, self.stroke_color.blue, self.stroke_color.alpha],
            # "stroke-width": self.stroke_width,
            # "stroke-dasharray": self.stroke_dash,
            # "location": {
            #     "x": self._location_x,
            #     "y": self._location_y
            # }
        }
        if self.fill_color != Color("white"):
            ds['fill'] = [self.fill_color.red, self.fill_color.green, self.fill_color.blue, self.fill_color.alpha]
        if self.opacity != 1.0:
            ds['opacity'] = self.opacity
        if self.stroke_color != Color("white"):
            ds['stroke'] = [self.stroke_color.red, self.stroke_color.green, self.stroke_color.blue, self.stroke_color.alpha]
        if self.stroke_width != 1.0:
            ds['stroke-width'] = self.stroke_width
        if self.stroke_dash != 1:
            ds['stoke-dasharray'] = self.stroke_dash
        if self._location_x != 0.0 or self._location_y != 0.0:
            ds['location'] = {
                "x": self._location_x,
                "y": self._location_y
            }
        return ds





class MovieActorWikiData:

    def __init__(self):
        self._movie_uri = ""
        self._actor_uri = ""
        self._movie_name = ""
        self._actor_name = ""

    @property
    def movie_uri(self):
        return self._movie_uri

    @movie_uri.setter
    def movie_uri(self, uri):
        self._movie_uri = uri

    @property
    def actor_uri(self):
        return self._actor_uri

    @actor_uri.setter
    def actor_uri(self, uri):
        self._actor_uri = uri

    @property
    def movie_name(self):
        return self._movie_name

    @movie_name.setter
    def movie_name(self, name):
        self._movie_name = name

    @property
    def actor_name(self):
        return self._actor_name

    @actor_name.setter
    def actor_name(self, name):
        self._actor_name = name




class CancerIncidence:

    def __init__(self):
        self.loc = []
        self.year = 0
        self.count = 0
        self.population = 0
        self.gender = ""
        self.race = ""
        self.event_type = ""
        self.affected_area = ""
        self.ageAdjustedRate = 0.0
        self.crudeRate = 0.0
        self.loc.append(0.0)
        self.loc.append(0.0)
        self.ageAdjustedRateci= []
        self.crudeRateci = []
        self.ageAdjustedRateci.append(0.0)
        self.ageAdjustedRateci.append(0.0)
        self.crudeRateci.append(0.0)
        self.crudeRateci.append(0.0)

    def get_age_adjusted_rate(self):
        return self.ageAdjustedRate

    def set_age_adjusted_rate(self, arr):
        self.ageAdjustedRate = arr

    def get_age_adjusted_ci_lower(self):
        return self.ageAdjustedRateci[0]

    def set_age_adjusted_ci_lower(self, ci_l):
        self.ageAdjustedRateci[0] = ci_l

    def get_age_adjusted_ci_upper(self):
        return self.ageAdjustedRateci[1]

    def set_age_adjusted_ci_upper(self, ci_u):
        self.ageAdjustedRateci[1] = ci_u

    def get_crude_rate(self):
        return self.crudeRate

    def set_crude_rate(self, cr):
        self.crudeRate = cr

    def get_crude_rate_ci_lower(self):
        return self.crudeRateci[0]

    def set_crude_rate_ci_lower(self, cr_l):
        self.crudeRateci[0] = cr_l

    def get_crude_rate_ci_upper(self):
        return self.crudeRateci[1]

    def set_crude_rate_ci_upper(self, cr_l):
        self.crudeRateci[1] = cr_l

    def get_year(self):
        return self.year

    def set_year(self, year):
        self.year = year

    def get_gender(self):
        return self.gender

    def set_gender(self, gender):
        self.gender = gender

    def get_race(self):
        return self.race

    def set_race(self, race):
        self.race = race

    def get_event_type(self):
        return self.event_type

    def set_event_type(self, et):
        self.event_type = et

    def get_population(self):
        return self.population

    def set_population(self, pop):
        self.population = pop

    def get_affected_area(self):
        return self.affected_area

    def set_affected_area(self, area):
        self.affected_area = area

    def get_count(self):
        return self.count

    def set_count(self,count):
        self.count = count

    def get_location_x(self):
        return self.loc[0]

    def set_location_x(self, x):
        self.loc[0] = x

    def get_location_y(self):
        return self.loc[1]

    def set_location_y(self, y):
        self.loc[1] = y
